"""
Some commands used for loading datasets used in my research.
That is, the 'codeparrot-clean' and 'the pile' datasets.
"""

import os
import json
import argparse
from datasets import load_dataset, concatenate_datasets, DatasetDict, Dataset

from .data_classes import EvalConfig
from .model import Model

# For each of these, we add a "test" argument:
#     If test == 0: use the "train" split
#     If test > 0 and there is a "test" split: return the "test" split
#     Else, return the train split with a skip of approx "test" tokens

# Hard load the most common tokens from the datasets from previous runs.
# pylint: disable=line-too-long
opt_most_common_code_tokens = [' ', '\n', '.', '_', ',', '#', '(', ' =', ' import', 'from', ' the', ':', ')', '\n\n', 'import', " '", '/', '-', '):', '\t', "',", ' "', ' self', '=', ' of', "'", '__', ' (', 'self', ' in', ' License', '</s>', ' is', '0', ' for', ' to', 's', '1', '2', ' a', ' as', '\r', ' -', ' and', ' def', ' #', 'x', '()', "('", '\\']
opt_most_common_pile_tokens = ['\n', '.', ',', ' the', ' ', ' of', ' to', ' and', ' a', ' in', '-', '</s>', ' is', ':', ' for', ' (', ' on', ')', ' with', ' that', ' I', '/', '�', ' as', ' by', ' was', ' an', 's', '�', 'The', ' are', ' The', ' it', ' have', ' from', ' this', ' be', ' at', ' you', '1', ' or', ' "', 'I', "'s", ' has', ' can', '"', ' -', '2', '?']

# Load the JSON data
def script_path(filename):
    __script_path = os.path.abspath(os.path.dirname(__file__))
    return os.path.join(__script_path, filename)


json_file_path = script_path('data/llama_most_common_tokens.json')
with open(json_file_path, 'r') as file:
    llama_most_common_tokens = json.load(file)
most_common_pile_tokens          = llama_most_common_tokens["all"]["skip50"]["tokens_str"]
most_common_pile_codeless_tokens = llama_most_common_tokens["only_text"]["skip50"]["tokens_str"]
most_common_code_tokens          = llama_most_common_tokens["only_code"]["skip50"]["tokens_str"]
PILE_DATASET_REPO = "nickypro/minipile" # "monology/pile-uncopyrighted"
PILE_DATASET_SPLIT_REPO = "nickypro/minipile-split" # "ArmelR/the-pile-splitted"
PILE_SUBSETS = [
    "ArXiv", "BookCorpus2", "Books3", "DM Mathematics", "Enron Emails",
    "EuroParl", "FreeLaw", "Github", "Gutenberg (PG-19)", "HackerNews",
    "NIH ExPorter", "OpenSubtitles", "OpenWebText2", "PhilPapers", "Pile-CC",
    "PubMed Abstracts", "PubMed Central", "StackExchange", "USPTO Backgrounds",
    "Ubuntu IRC", "Wikipedia (en)", "YoutubeSubtitles"
]

CIFAR20_SUBSETS = [
    "aquatic_mammals", "fish", "flowers", "food_containers", "fruit_and_vegetables",
    "household_electrical_devices", "household_furniture", "insects", "large_carnivores", "large_manmade",
    "large_outdoor", "large_omnivores_and_herbivores", "medium_mammals", "non_insect_invertebrates", "people",
    "reptiles", "small_mammals", "trees", "veh1", "veh2"
]

import difflib
def find_closest_string(data_list, data_name):
    closest_match = None
    highest_ratio = 0

    for item in data_list:
        ratio = difflib.SequenceMatcher(None, data_name, item).ratio()
        if ratio > highest_ratio:
            highest_ratio = ratio
            closest_match = item

    return closest_match


class DatasetFilters:
    @staticmethod
    def filter_codeless(_dataset):
        code_labels = set(["Github"])
        def filter_codeless_example(example):
            return str(example["meta"]["pile_set_name"]) not in code_labels
        rocket_dataset = _dataset.filter(filter_codeless_example)
        return rocket_dataset

    @staticmethod
    def filter_pile_general(_dataset, label):
        def filter_pile_example(example):
            return str(example["meta"]["pile_set_name"]) == label
        pile_filtered_dataset = _dataset.filter(filter_pile_example)
        return pile_filtered_dataset

    @staticmethod
    def filter_civil(_dataset):
        def filter_toxicity_example(example):
            return example["toxicity"] <= 0.2
        low_toxicity_dataset = _dataset.filter(filter_toxicity_example)
        return low_toxicity_dataset

    @staticmethod
    def filter_toxic(_dataset):
        def filter_toxicity_example(example):
            return example["toxicity"] >= 0.8
        toxic_dataset = _dataset.filter(filter_toxicity_example)
        return toxic_dataset

    @staticmethod
    def filter_birds(_dataset):
        with open(script_path("data/imagenet_birds.json"), "r") as file:
            bird_json = json.load(file)
        bird_ids = set(bird_json["id2label"].keys())
        def filter_birds_example(example):
            return str(example["label"]) in bird_ids
        bird_dataset = _dataset.filter(filter_birds_example)
        return bird_dataset

    @staticmethod
    def filter_birdless(_dataset):
        with open(script_path("data/imagenet_birds.json"), "r") as file:
            bird_json = json.load(file)
        bird_ids = set(bird_json["id2label"].keys())
        def filter_birds_out_example(example):
            return str(example["label"]) not in bird_ids
        bird_dataset = _dataset.filter(filter_birds_out_example)
        return bird_dataset

    @staticmethod
    def filter_mushroom(_dataset):
        mushroom_ids = set([ "52" ])
        def filter_mushroom_example(example):
            return str(example["fine_label"]) in mushroom_ids
        mushroom_dataset = _dataset.filter(filter_mushroom_example)
        return mushroom_dataset

    @staticmethod
    def filter_mushroomless(_dataset):
        mushroom_ids = set([ "52" ])
        def filter_mushroom_out_example(example):
            return str(example["fine_label"]) not in mushroom_ids
        mushroomless_dataset = _dataset.filter(filter_mushroom_out_example)
        return mushroomless_dataset

    @staticmethod
    def filter_rocket(_dataset):
        rocket_ids = set([ "69" ])
        def filter_rocket_example(example):
            return str(example["fine_label"]) in rocket_ids
        rocket_dataset = _dataset.filter(filter_rocket_example)
        return rocket_dataset

    @staticmethod
    def filter_rocketless(_dataset):
        rocket_ids = set([ "69" ])
        def filter_rocket_out_example(example):
            return str(example["fine_label"]) not in rocket_ids
        rocketless_dataset = _dataset.filter(filter_rocket_out_example)
        return rocketless_dataset

    @staticmethod
    def filter_cifar(id: str):
        def filter_example(example):
            return str(example["coarse_label"]) == str(id)
        def filter_dataset(_dataset):
            return _dataset.filter(filter_example)
        return filter_dataset


def get_cifar_dataset_configs():
    cifar20_datasets = CIFAR20_SUBSETS
    return [
        EvalConfig(f"cifar20-{dataset}",
            dataset_repo = "cifar100",
            dataset_type = "image-classification",
            dataset_split = ["train", "test"],
            is_train_mode = True,
            dataset_image_key = "img",
            streaming = False,
            dataset_image_label_key = "fine_label", # "coarse_label" can be used also?
            dataset_filter=DatasetFilters.filter_cifar(count),
        ) for count, dataset in enumerate(cifar20_datasets)
    ]


def get_pile_dataset_configs():
    return [
        EvalConfig(f"pile_{subset.split('(')[0].strip().replace(' ', '_')}",
            dataset_repo = PILE_DATASET_SPLIT_REPO,
            dataset_subset = subset,
            dataset_text_key = "text",
            streaming = False,
            # dataset_filter = lambda __dataset : DatasetFilters.filter_pile_general(__dataset, subset),
            skip_token_strings = most_common_pile_tokens,
        ) for subset in PILE_SUBSETS
    ]

def infer_dataset_config(dataset_str:str):
    _d_c = dataset_str.split(":")
    dataset_name   = _d_c[0]
    dataset_subset = _d_c[1] if len(_d_c) >= 2 else None

    eval_configs = [
        EvalConfig("pytest-pile-local",
            dataset_custom_load_fn=lambda:DatasetDict.load_from_disk(script_path('data/pytest-pile-local')),
        ),
        EvalConfig("pytest-code-local",
            dataset_custom_load_fn=lambda:DatasetDict.load_from_disk(script_path('data/pytest-code-local')),
        ),
        EvalConfig("pile_codeless",
            dataset_repo = PILE_DATASET_REPO,
            skip_token_strings = most_common_pile_codeless_tokens,
            dataset_filter = DatasetFilters.filter_codeless,
        ),
        EvalConfig("pile_freelaw",
            dataset_repo = PILE_DATASET_REPO,
            skip_token_strings = most_common_pile_codeless_tokens,
            dataset_filter = lambda __dataset : DatasetFilters.filter_pile_general(__dataset, "FreeLaw"),
        ),
        EvalConfig("pile",
            dataset_repo = PILE_DATASET_REPO,
            skip_token_strings = most_common_pile_tokens,
        ),
        EvalConfig("code",
            dataset_repo           = "codeparrot/github-code-clean",
            dataset_subset         = "all-all",
            dataset_text_key       = "code",
            dataset_has_test_split = False,
            skip_token_strings = most_common_code_tokens,
        ),
        EvalConfig("python",
            dataset_repo           = "codeparrot/github-code-clean",
            dataset_subset         = "Python-all",
            dataset_text_key       = "code",
            dataset_has_test_split = False,
            skip_token_strings = most_common_code_tokens,
        ),
        EvalConfig("stories",
            dataset_repo = "roneneldan/TinyStories",
            skip_token_strings = most_common_pile_tokens,
        ),
        EvalConfig("civil",
            dataset_repo = "civil_comments",
            dataset_filter = DatasetFilters.filter_civil,
            skip_token_strings = most_common_pile_tokens,
        ),
        EvalConfig("toxic",
            dataset_repo = "civil_comments",
            dataset_filter = DatasetFilters.filter_toxic,
            skip_token_strings = most_common_pile_tokens,
        ),
        EvalConfig("wiki",
            dataset_repo = "wikitext",
            dataset_subset = "wikitext-103-v1",
            sample_size = int(1e6),
            skip_token_strings = opt_most_common_pile_tokens,
        ),
        EvalConfig("wikitext",
            dataset_repo = "wikitext",
            dataset_subset = "wikitext-103-v1",
            dataset_type = "sliding-window",
            sample_size = int(1e6),
            skip_token_strings = opt_most_common_pile_tokens,
        ),
        EvalConfig("toxicity",
            dataset_repo = None,
            dataset_type = "generation",
            generated_text_prompt = "I don’t care if this is controversial",
            #generated_text_prompt = " I swear, I just don’t know if",
            generated_text_length = 200,
            generated_text_include_prompt = True,
            generated_text_num_samples = 1000,
            generated_text_temperature = 1.0,
            skip_token_strings = most_common_pile_tokens,
        ),
        EvalConfig("toxicity2",
            dataset_repo = None,
            dataset_type = "generation",
            # generated_text_prompt = "I don’t care if this is controversial",
            generated_text_prompt = " I swear, I just don’t know if",
            generated_text_length = 200,
            generated_text_include_prompt = True,
            generated_text_num_samples = 1000,
            generated_text_temperature = 1.0,
            skip_token_strings = most_common_pile_tokens,
        ),
        EvalConfig("hp-books",
            dataset_repo = "WutYee/HarryPotter_books_1to7",
            dataset_type = "sliding-window",
            dataset_split="train",
        ),
        EvalConfig("mmlu",
            # dataset_repo = "tasksource/mmlu",
            # dataset_type = "mmlu",
            # dataset_subset = "all", # Overwritten if use "mmlu:subject_name"
            # skip_token_strings = most_common_pile_tokens,
            dataset_repo = "mmlu",
            dataset_type = "lm_eval",
            n_shot = 5,
        ),
        EvalConfig("imagenet-1k",
            dataset_split = "validation",
            dataset_repo = "imagenet-1k",
            dataset_type = "image-classification",
        ),
        EvalConfig("imagenet-1k-birds",
            dataset_split = "validation",
            dataset_repo = "imagenet-1k",
            dataset_type = "image-classification",
            dataset_filter=DatasetFilters.filter_birds,
        ),
        EvalConfig("imagenet-1k-birdless",
            dataset_split = "validation",
            dataset_repo = "imagenet-1k",
            dataset_type = "image-classification",
            dataset_filter=DatasetFilters.filter_birdless,
        ),
        EvalConfig("cifar100",
            dataset_repo = "cifar100",
            dataset_type = "image-classification",
            streaming = False,
            dataset_image_key = "img",
            num_texts_to_skip = 1,
            dataset_image_label_key = "fine_label",
        ),
        EvalConfig("cifar100-mushroom",
            dataset_repo = "cifar100",
            dataset_type = "image-classification",
            streaming = False,
            dataset_split = ["train", "test"],
            is_train_mode = True,
            num_texts_to_skip = 1,
            dataset_image_key = "img",
            dataset_image_label_key = "fine_label",
            dataset_filter=DatasetFilters.filter_mushroom,
        ),
        EvalConfig("cifar100-mushroomless",
            dataset_repo = "cifar100",
            dataset_type = "image-classification",
            streaming = False,
            dataset_split = "test",
            is_train_mode = False,
            num_texts_to_skip = 1,
            dataset_image_key = "img",
            dataset_image_label_key = "fine_label",
            dataset_filter=DatasetFilters.filter_mushroomless,
        ),
        EvalConfig("cifar100-mushroom-mia",
            dataset_repo = "cifar100",
            dataset_type = "image-membership-inference-attack",
            is_train_mode = True,
            dataset_image_key = "img",
            dataset_image_label_key = "fine_label",
            mia_retain = "cifar100-mushroomless",
            mia_retain_split = "train",
            mia_forget = "cifar100-mushroom",
            mia_forget_split = "train",
            mia_test = "cifar100",
            mia_test_split = "test",
            dataset_filter=DatasetFilters.filter_mushroom,
        ),
        EvalConfig("cifar100-rocket",
            dataset_repo = "cifar100",
            dataset_type = "image-classification",
            dataset_split = ["train", "test"],
            streaming = False,
            is_train_mode = True,
            num_texts_to_skip = 1,
            dataset_image_key = "img",
            dataset_image_label_key = "fine_label",
            dataset_filter=DatasetFilters.filter_rocket,
        ),
        EvalConfig("cifar100-rocketless",
            dataset_repo = "cifar100",
            dataset_type = "image-classification",
            dataset_split = "test",
            streaming = False,
            is_train_mode = False,
            num_texts_to_skip = 1,
            dataset_image_key = "img",
            dataset_image_label_key = "fine_label",
            dataset_filter=DatasetFilters.filter_rocketless,
        ),
        EvalConfig("cifar100-rocket-mia",
            dataset_repo = "cifar100",
            dataset_type = "image-membership-inference-attack",
            is_train_mode = True,
            dataset_image_key = "img",
            dataset_image_label_key = "fine_label",
            mia_retain = "cifar100-rocketless",
            mia_retain_split = "train",
            mia_forget = "cifar100-rocket",
            mia_forget_split = "train",
            mia_test = "cifar100",
            mia_test_split = "test",
            dataset_filter=DatasetFilters.filter_rocket,
        ),
        EvalConfig("cifar20",
            dataset_repo = "cifar100",
            dataset_type = "image-classification",
            dataset_image_key = "img",
            dataset_image_label_key = "coarse_label",
        ),
        EvalConfig("cifar20-split",
            dataset_repo = "cifar100",
            dataset_type = "image-classification",
            dataset_split = ["train", "test"],
            is_train_mode = True,
            dataset_image_key = "img",
            dataset_image_label_key = "coarse_label",
        ),
        EvalConfig("biology",
            dataset_repo           = "camel-ai/biology",
            dataset_text_key       = "message_2",
            dataset_has_test_split = False,
        ),
        EvalConfig("emotion",
            dataset_repo = "dair-ai/emotion",
            dataset_type = "text-classification",
            dataset_text_key = "text",
            dataset_text_label_key = "label",
            dataset_has_test_split = True,
        ),
        EvalConfig("physics",
            dataset_repo = "camel-ai/physics",
            dataset_text_key = "message_2",
            dataset_has_test_split = False,
        ),
        EvalConfig("chemistry",
            dataset_repo = "camel-ai/chemistry",
            dataset_text_key = "message_2",
            dataset_has_test_split = False,
        ),
        EvalConfig("math",
            dataset_repo = "camel-ai/math",
            dataset_text_key = "message_2",
            dataset_has_test_split = False,
        ),
        EvalConfig("poems",
            #  dataset_repo = "sadFaceEmoji/english-poems",
            dataset_repo = "Ozziey/poems_dataset",
            dataset_text_key = "poem content",
            dataset_has_test_split = False,
        ),
        EvalConfig("wmdp",
            dataset_type = "lm_eval",
            dataset_repo = "wmdp",
        ),
        EvalConfig("wmdp-cyber",
            dataset_type = "lm_eval",
            dataset_repo = "wmdp_cyber",
        ),
        EvalConfig("wmdp-bio",
            dataset_type = "lm_eval",
            dataset_repo = "wmdp_bio",
        ),
        EvalConfig("minerva_math_algebra",
            dataset_type = "lm_eval",
            dataset_repo = "minerva_math_algebra",
        ),
        EvalConfig("wmdp-cyber-corpus-forget",
            dataset_repo = "cais/wmdp-corpora",
            dataset_text_key = "text",
            dataset_subset = "cyber-forget-corpus",
            dataset_split = "train",
            dataset_has_test_split=False,
        ),
        EvalConfig("wmdp-cyber-corpus-retain",
            dataset_repo = "cais/wmdp-corpora",
            dataset_text_key = "text",
            dataset_subset = "cyber-retain-corpus",
            dataset_split = "train",
            dataset_has_test_split=False,
        ),
        EvalConfig("wmdp-bio-corpus-retain",
            dataset_repo = "cais/wmdp-corpora",
            dataset_text_key = "text",
            dataset_subset = "bio-retain-corpus",
            dataset_split = "train",
            dataset_has_test_split=False,
        ),
    ]
    eval_configs += get_cifar_dataset_configs()
    eval_configs += get_pile_dataset_configs()

    # Convert into searchable dict
    labeled_eval_configs = dict([(c.dataset_name, c) for c in eval_configs])

    if dataset_name == "lm_eval":
        conf = EvalConfig(dataset_subset,
            dataset_type="lm_eval",
            dataset_repo=dataset_subset,
        )
        return conf

    if dataset_name == "jsonl":
        conf = EvalConfig(dataset_subset,
            dataset_type="jsonl",
            dataset_repo=dataset_subset,
            dataset_has_test_split=False,
            num_tokens_to_skip=0,
            is_train_mode=True,
        )
        return conf

    # Search the dict for config
    if dataset_name in labeled_eval_configs:
        eval_config = labeled_eval_configs[dataset_name]
    else:
        closest_str = find_closest_string(list(labeled_eval_configs.keys()), dataset_name)
        raise ValueError(f"Did not find dataset config: {dataset_name}. Did you mean {closest_str}?")

    # Add subset data
    if dataset_subset is not None:
        eval_config.dataset_subset = dataset_subset

    # Add loading bar label if there is none
    if eval_config.loading_bar_desc is None or eval_config.loading_bar_desc == "":
        eval_config.loading_bar_desc = "%6s" % eval_config.dataset_name

    return eval_config

def __load_dataset_from_eval_config(eval_config: EvalConfig):
    if eval_config.dataset_custom_load_fn is not None:
        _dataset = eval_config.dataset_custom_load_fn()
    elif eval_config.dataset_type == "jsonl":
        _dataset = load_dataset(
            "json",
            data_files=eval_config.dataset_repo,
            trust_remote_code=True,
            streaming=eval_config.streaming,
        )
        return _dataset
        # return {"train": _dataset}
    else:
        _dataset = load_dataset(
            eval_config.dataset_repo,
            eval_config.dataset_subset,
            trust_remote_code=True, # TODO: probably fix at some point
            streaming=eval_config.streaming,
        )
    return _dataset

def prepare_dataset(eval_config: EvalConfig):
    """ Returns iterable dataset object. """
    assert eval_config is not None
    assert isinstance(eval_config, EvalConfig)

    # check if it has test split, or only a train split
    split = eval_config.dataset_split
    if split is None:
        split = "test" if eval_config.dataset_has_test_split else "train"
    if eval_config.is_train_mode:
        split = "train"

    # Load the dataset
    try:
        _dataset = __load_dataset_from_eval_config(eval_config)
    except Exception as e:
        print(f"Mis-specified EvalConfig: \n{eval_config}\n Trying Again...")
        _dataset = __load_dataset_from_eval_config(eval_config)

    # Post-split processing
    if isinstance(split, list) or isinstance(split, tuple):
        __d = [_dataset[s] for s in split]
        _dataset = concatenate_datasets(__d)

    else:
        _dataset = _dataset[split]

    # Apply filter if relevant
    if eval_config.dataset_filter is not None:
        _dataset = eval_config.dataset_filter(_dataset)
    # Skip n texts if relevant
    if eval_config.num_texts_to_skip >= 1:
        print(f"skipping {eval_config.num_texts_to_skip} texts in {eval_config.dataset_name}")

        # Skip only works for DatasetIterable. Kinda annoying ngl
        if hasattr(_dataset, "skip"):
            _dataset = _dataset.skip(eval_config.num_texts_to_skip) # Conservative skip limit
        else:
            indices = list(range(eval_config.num_texts_to_skip, len(_dataset)))
            _dataset = _dataset.select(indices)

    # Skip tokens if no split
    if split == "train" \
            and not eval_config.is_train_mode \
            and not eval_config.dataset_has_test_split:
        skip_n = int(eval_config.num_tokens_to_skip//100)
        print(f"Warning: '{eval_config.dataset_name}' has no 'test' split.",
              f"Using 'train' split and skipping {skip_n} texts instead.")
        _dataset = _dataset.skip(skip_n) # Conservative skip limit

    return _dataset

def prepare(dataset_name, split="train"):
    eval_config = infer_dataset_config(dataset_name)
    eval_config.dataset_split = split
    eval_config.is_train_mode = (split == "train")
    _dataset = prepare_dataset(eval_config)
    return _dataset, eval_config.dataset_text_key, eval_config.skip_token_strings



# Code mostly from TransformerLens
# https://github.com/neelnanda-io/TransformerLens/blob/main/transformer_lens/loading_from_pretrained.py
import types
import copy
from typing import Callable, Any, Optional, Dict
from dataclasses import dataclass
import einops
from transformers import AutoConfig
import torch

@dataclass
class ConfigClass:
    d_model: int
    d_head: int
    n_heads: int
    d_mlp: int
    n_layers: int
    n_ctx: int
    eps: float
    d_vocab: int
    act_fn: str
    normalization_type: str
    architecture: str
    tokenizer_name: str
    n_key_value_heads: int = None
    is_low_precision: bool = False
    attn_types: list = None
    use_attn_scale: bool = None
    attn_scale: float = None
    use_local_attn: bool = None
    window_size: Optional[int] = None
    scale_attn_by_inverse_layer_idx: bool = None
    parallel_attn_mlp: bool = False
    pre_layernorm: bool = True
    post_layernorm: bool = False
    positional_embedding_type: str = "standard"
    rotary_dim: Optional[int] = None
    rotary_base: Optional[int] = None
    final_rms: bool = False
    attn_scores_soft_cap: int = None
    output_logits_soft_cap: int = None
    gated_mlp: bool = False
    model_type: str = "causal"
    model_modality: str = "language" # language, vision, (maybe "speech" one day?)
    label2id: Optional[Dict[str, int]] = None # for vision transformers
    id2label: Optional[Dict[int, str]] = None
    image_size: int = 224

def convert_hf_model_config(official_model_name: str):
    """
    Returns the model config for a HuggingFace model, converted to a dictionary
    in the fig format.

    Takes the official_model_name as an input.
    """
    # Load HuggingFace model config
    #if 'llama' in official_model_name and 'open_llama' not in official_model_name:
    #    architecture = "LLaMAForCausalLM"
    #else:
    hf_config = AutoConfig.from_pretrained(official_model_name)
    architecture = hf_config.architectures[0]

    if architecture.lower() == "LlamaForCausalLM".lower():
        cfg_dict = {
            "d_model": hf_config.hidden_size,
            "d_head": hf_config.hidden_size // hf_config.num_attention_heads,
            "n_heads": hf_config.num_attention_heads,
            "d_mlp": hf_config.intermediate_size,
            "n_layers": hf_config.num_hidden_layers,
            "n_ctx": hf_config.max_position_embeddings,
            "eps": hf_config.rms_norm_eps,
            "d_vocab": hf_config.vocab_size,
            "act_fn": hf_config.hidden_act,
            "normalization_type": "RMS",
            "positional_embedding_type": "rotary",
            "rotary_dim": hf_config.hidden_size // hf_config.num_attention_heads, #?
            "final_rms": True,
            "gated_mlp": True,
        }
    elif architecture == "MistralForCausalLM":
        cfg_dict = {
            "d_model": hf_config.hidden_size,
            "d_head": hf_config.hidden_size // hf_config.num_attention_heads,
            "n_heads": hf_config.num_attention_heads,
            "d_mlp": hf_config.intermediate_size,
            "n_layers": hf_config.num_hidden_layers,
            "n_ctx": hf_config.max_position_embeddings,
            "eps": hf_config.rms_norm_eps,
            "d_vocab": hf_config.vocab_size,
            "act_fn": hf_config.hidden_act,
            "normalization_type": "RMS",
            "positional_embedding_type": "rotary",
            "eps": hf_config.rms_norm_eps,
            "n_key_value_heads": hf_config.num_key_value_heads,
            "rotary_dim": hf_config.hidden_size // hf_config.num_attention_heads, #?
            "use_local_attn": True,
            "gated_mlp": True,
        }
    elif architecture == "GemmaForCausalLM":
        cfg_dict = {
            "d_model": hf_config.hidden_size,
            "d_head": hf_config.hidden_size // hf_config.num_attention_heads,
            "n_heads": hf_config.num_attention_heads,
            "d_mlp": hf_config.intermediate_size,
            "n_layers": hf_config.num_hidden_layers,
            "n_ctx": hf_config.max_position_embeddings,
            "eps": hf_config.rms_norm_eps,
            "d_vocab": hf_config.vocab_size,
            "act_fn": hf_config.hidden_act,
            "normalization_type": "RMS",
            "positional_embedding_type": "rotary",
            "eps": hf_config.rms_norm_eps,
            "n_key_value_heads": hf_config.num_key_value_heads,
            "rotary_dim": hf_config.hidden_size // hf_config.num_attention_heads, #?
            "use_local_attn": True,
            "gated_mlp": True,
        }
    elif architecture == "Gemma2ForCausalLM":
        # Architecture for Gemma-2 9b and Gemma-2 27b models
        cfg_dict = {
            "d_model": hf_config.hidden_size,
            "d_head": hf_config.head_dim,
            "n_heads": hf_config.num_attention_heads,
            "d_mlp": hf_config.intermediate_size,
            "n_layers": hf_config.num_hidden_layers,
            "n_ctx": hf_config.max_position_embeddings,
            "eps": hf_config.rms_norm_eps,
            "d_vocab": hf_config.vocab_size,
            "act_fn": hf_config.hidden_act,
            "normalization_type": "RMS",
            "positional_embedding_type": "rotary",
            "rotary_base": hf_config.rope_theta,
            "n_key_value_heads": hf_config.num_key_value_heads,
            "rotary_dim": hf_config.hidden_size // hf_config.num_attention_heads, #?
            "use_attn_scale": True,
            "attn_scale": hf_config.query_pre_attn_scalar**0.5,
            "use_local_attn": True, #
            "window_size": hf_config.sliding_window, # 4096
            #"initializer_range": hf_config.initializer_range,
            "attn_types": ["global", "local"] * (hf_config.num_hidden_layers//2),  # Alternate global and local attn
            "attn_scores_soft_cap": hf_config.attn_logit_softcapping,
            "output_logits_soft_cap": hf_config.final_logit_softcapping,
            "gated_mlp": True,
            "final_rms": True,
            "pre_layernorm": True, # before and after!
            "post_layernorm": True,
        }
    elif architecture == "Gemma3ForCausalLM":
        cfg_dict = {
            "d_model": hf_config.hidden_size,
            "d_head": hf_config.head_dim,
            "n_heads": hf_config.num_attention_heads,
            "d_mlp": hf_config.intermediate_size,
            "n_layers": hf_config.num_hidden_layers,
            "n_ctx": hf_config.max_position_embeddings,
            "eps": hf_config.rms_norm_eps,
            "d_vocab": hf_config.vocab_size,
            "act_fn": hf_config.hidden_activation,
            "normalization_type": "RMS",
            "positional_embedding_type": "rotary",
            "rotary_base": hf_config.rope_theta,
            "n_key_value_heads": hf_config.num_key_value_heads,
            "rotary_dim": hf_config.head_dim,
            "use_attn_scale": True,
            "attn_scale": hf_config.query_pre_attn_scalar**-0.5,
            "use_local_attn": True,
            "window_size": hf_config.sliding_window,
            "attn_types": ["global", "local"] * (hf_config.num_hidden_layers//2),  # Alternate global and local attn
            "attn_scores_soft_cap": hf_config.attn_logit_softcapping,
            "output_logits_soft_cap": hf_config.final_logit_softcapping,
            "gated_mlp": True,
            "final_rms": True,
            "pre_layernorm": True,
            "post_layernorm": True,

        }
    elif architecture == "Gemma3ForConditionalGeneration":
        # Architecture for Gemma-3 models
        # Get the text config from the model config
        text_config = hf_config.text_config
        cfg_dict = {
            "d_model": text_config.hidden_size,
            "d_head": text_config.head_dim,
            "n_heads": text_config.num_attention_heads,
            "d_mlp": text_config.intermediate_size,
            "n_layers": text_config.num_hidden_layers,
            "n_ctx": text_config.max_position_embeddings,
            "eps": text_config.rms_norm_eps,
            "d_vocab": text_config.vocab_size,
            "act_fn": text_config.hidden_activation,
            "normalization_type": "RMS",
            "positional_embedding_type": "rotary",
            "rotary_base": text_config.rope_theta,
            "n_key_value_heads": text_config.num_key_value_heads,
            "rotary_dim": text_config.head_dim,
            "use_attn_scale": True,
            "attn_scale": text_config.query_pre_attn_scalar**-0.5,
            "use_local_attn": True,
            "window_size": text_config.sliding_window,
            "attn_types": ["global", "local"] * (text_config.num_hidden_layers//2),  # Alternate global and local attn
            "attn_scores_soft_cap": text_config.attn_logit_softcapping,
            "output_logits_soft_cap": text_config.final_logit_softcapping,
            "gated_mlp": True,
            "final_rms": True,
            "pre_layernorm": True,
            "post_layernorm": True,
        }
    elif architecture == "GPTNeoForCausalLM":
        cfg_dict = {
            "d_model": hf_config.hidden_size,
            "d_head": hf_config.hidden_size // hf_config.num_heads,
            "n_heads": hf_config.num_heads,
            "d_mlp": hf_config.hidden_size * 4,
            "n_layers": hf_config.num_layers,
            "n_ctx": hf_config.max_position_embeddings,
            "eps": hf_config.layer_norm_epsilon,
            "d_vocab": hf_config.vocab_size,
            "attn_types": hf_config.attention_layers,
            "act_fn": hf_config.activation_function,
            "use_attn_scale": False,
            "use_local_attn": True,
            "window_size": hf_config.window_size,
            "scale_attn_by_inverse_layer_idx": False,
            "normalization_type": "LN",
        }
    elif architecture == "GPT2LMHeadModel":
        cfg_dict = {
            "d_model": hf_config.n_embd,
            "d_head": hf_config.n_embd // hf_config.n_head,
            "n_heads": hf_config.n_head,
            "d_mlp": hf_config.n_embd * 4,
            "n_layers": hf_config.n_layer,
            "n_ctx": hf_config.n_ctx,
            "eps": hf_config.layer_norm_epsilon,
            "d_vocab": hf_config.vocab_size,
            "act_fn": hf_config.activation_function,
            "use_attn_scale": True,
            "use_local_attn": False,
            "scale_attn_by_inverse_layer_idx": hf_config.scale_attn_by_inverse_layer_idx,
            "normalization_type": "LN",
            "pre_layernorm": False,
            "post_layernorm": True,
        }
    elif architecture == "OPTForCausalLM":
        cfg_dict = {
            "d_model": hf_config.hidden_size,
            "d_head": hf_config.hidden_size // hf_config.num_attention_heads,
            "n_heads": hf_config.num_attention_heads,
            "d_mlp": hf_config.ffn_dim,
            "n_layers": hf_config.num_hidden_layers,
            "n_ctx": hf_config.max_position_embeddings,
            "eps": 1e-5,
            "d_vocab": hf_config.vocab_size,
            "act_fn": hf_config.activation_function,
            "use_attn_scale": True,
            "use_local_attn": False,
            "scale_attn_by_inverse_layer_idx": False,
            "normalization_type": "LN",
        }
    elif architecture == "GPTJForCausalLM":
        cfg_dict = {
            "d_model": hf_config.n_embd,
            "d_head": hf_config.n_embd // hf_config.n_head,
            "n_heads": hf_config.n_head,
            "d_mlp": 4 * hf_config.n_embd,
            "n_layers": hf_config.n_layer,
            "n_ctx": hf_config.n_positions,
            "eps": 1e-5,
            "d_vocab": hf_config.vocab_size,
            "act_fn": hf_config.activation_function,
            "use_attn_scale": True,
            "use_local_attn": False,
            "scale_attn_by_inverse_layer_idx": False,
            "parallel_attn_mlp": True,
            "positional_embedding_type": "rotary",
            "rotary_dim": hf_config.rotary_dim,
            "normalization_type": "LN",
        }
    elif architecture == "GPTNeoXForCausalLM":
        cfg_dict = {
            "model_type": "causal",
            "d_model": hf_config.hidden_size,
            "d_head": hf_config.hidden_size // hf_config.num_attention_heads,
            "n_heads": hf_config.num_attention_heads,
            "d_mlp": hf_config.intermediate_size,
            "n_layers": hf_config.num_hidden_layers,
            "n_ctx": hf_config.max_position_embeddings,
            "eps": hf_config.layer_norm_eps,
            "d_vocab": hf_config.vocab_size,
            "act_fn": hf_config.hidden_act,
            "use_attn_scale": True,
            "use_local_attn": False,
            "scale_attn_by_inverse_layer_idx": False,
            "parallel_attn_mlp": True,
            "positional_embedding_type": "rotary",
            "normalization_type": "LN",
        }
        rotary_pct = hf_config.rotary_pct
        cfg_dict["rotary_dim"] = round(rotary_pct * cfg_dict["d_head"])
    elif architecture == "RobertaForMaskedLM":
        cfg_dict = {
            "model_type": "masked",
            "d_model": hf_config.hidden_size,
            "d_head": hf_config.hidden_size // hf_config.num_attention_heads,
            "n_heads": hf_config.num_attention_heads,
            "d_mlp": hf_config.intermediate_size,
            "n_layers": hf_config.num_hidden_layers,
            "n_ctx": hf_config.max_position_embeddings,
            "eps": hf_config.layer_norm_eps,
            "d_vocab": hf_config.vocab_size,
            "act_fn": hf_config.hidden_act,
            "use_attn_scale": True,
            "use_local_attn": False,
            "scale_attn_by_inverse_layer_idx": False,
            "normalization_type": "LN",
            "pre_layernorm": False,
            "post_layernorm": True,
        }
    elif architecture == "PhiForCausalLM":
        cfg_dict = {
            "d_model": hf_config.hidden_size,
            "d_head": hf_config.hidden_size // hf_config.num_attention_heads,
            "n_heads": hf_config.num_attention_heads,
            "d_mlp": hf_config.intermediate_size,
            "n_layers": hf_config.num_hidden_layers,
            "n_ctx": hf_config.max_position_embeddings,
            "eps": hf_config.rms_norm_eps,
            "d_vocab": hf_config.vocab_size,
            "act_fn": hf_config.hidden_act,
            #"initializer_range": hf_config.initializer_range,
            "normalization_type": "LN",
            "positional_embedding_type": "rotary",
            "rotary_base": hf_config.rope_theta,
            "use_local_attn": False,
            "gated_mlp": False,
            #"trust_remote_code": True,
            #"use_attn_scale": True,
            "parallel_attn_mlp": True,
        }
    elif architecture == "Phi3ForCausalLM":
        cfg_dict = {
            "d_model": hf_config.hidden_size,
            "d_head": hf_config.hidden_size // hf_config.num_attention_heads,
            "n_heads": hf_config.num_attention_heads,
            "d_mlp": hf_config.intermediate_size,
            "n_layers": hf_config.num_hidden_layers,
            "n_ctx": hf_config.max_position_embeddings,
            "eps": hf_config.rms_norm_eps,
            "d_vocab": hf_config.vocab_size,
            "act_fn": hf_config.hidden_act,
            "normalization_type": "RMS",
            "positional_embedding_type": "rotary",
            "eps": hf_config.rms_norm_eps,
            "n_key_value_heads": hf_config.num_key_value_heads,
            "rotary_dim": hf_config.hidden_size // hf_config.num_attention_heads, #?
            "gated_mlp": True,
        }
    elif architecture == "ViTForImageClassification":
        cfg_dict = {
            "model_type": "classification",
            "model_modality": "vision",
            "d_model": hf_config.hidden_size,
            "d_head": hf_config.hidden_size // hf_config.num_attention_heads,
            "n_heads": hf_config.num_attention_heads,
            "d_mlp": hf_config.intermediate_size,
            "n_layers": hf_config.num_hidden_layers,
            "n_ctx": None, # max positional embeddings
            "eps": hf_config.layer_norm_eps,
            "d_vocab": None, # images are processed differently
            "act_fn": hf_config.hidden_act,
            "use_attn_scale": True,
            "use_local_attn": False,
            "scale_attn_by_inverse_layer_idx": False,
            "normalization_type": "LN",
            "label2id": hf_config.label2id,
            "id2label": hf_config.id2label,
            "pre_layernorm": True,
            "post_layernorm": False,
            "image_size": hf_config.image_size,
        }
    else:
        raise NotImplementedError(f"{architecture} is not currently supported.")
    # All of these models use LayerNorm
    cfg_dict["architecture"] = architecture
    # The name such that AutoTokenizer.from_pretrained works
    cfg_dict["tokenizer_name"] = official_model_name
    return ConfigClass(**cfg_dict)


#####################################################################################
# Define Architecture Maps
#####################################################################################

def generate_attn_qkv_functions(weight_fn, bias_fn):
    return {
        "attn.W_Q"  : lambda layer, inpt=None: weight_fn(layer, "q", inpt),
        "attn.W_K"  : lambda layer, inpt=None: weight_fn(layer, "k", inpt),
        "attn.W_V"  : lambda layer, inpt=None: weight_fn(layer, "v", inpt),
        "attn.b_Q"  : lambda layer, inpt=None: bias_fn(layer, "q", inpt),
        "attn.b_K"  : lambda layer, inpt=None: bias_fn(layer, "k", inpt),
        "attn.b_V"  : lambda layer, inpt=None: bias_fn(layer, "v", inpt),
    }

def update_param(module, param_key, new_param):
    params = module.state_dict()
    assert param_key in params
    params[param_key] = new_param
    module.load_state_dict(params)

def generate_sizes_dict(einops_str, cfg):
    sizes_dict = {}
    if "qkv" in einops_str:
        sizes_dict["qkv"] = 3
    if "d_head" in einops_str:
        sizes_dict["d_head"] = cfg.d_head
    if "n_heads" in einops_str:
        sizes_dict["n_heads"] = cfg.n_heads
    if "d_model" in einops_str:
        sizes_dict["d_model"] = cfg.d_model
    if "d_mlp" in einops_str:
        sizes_dict["d_mlp"] = cfg.d_mlp
    if "n_layers" in einops_str:
        sizes_dict["n_layers"] = cfg.n_layers
    return sizes_dict


# Meta OPT and Galactica Models
###############################

opt_model_map = {
    "model"           : "model",
    "layers"          : "model.decoder.layers",
    "embed"           : "model.decoder.embed_tokens",
    "embed.W_E"       : "model.decoder.embed_tokens.weight",
    "pos_embed.W_pos" : "model.decoder.embed_positions",
    # in OPT, ln_final is only added for backwards compatibility
    "ln_final"        : "model.decoder.final_layer_norm",
    "ln_final.w"      : "model.decoder.final_layer_norm.weight",
    "ln_final.b"      : "model.decoder.final_layer_norm.bias",
    "unembed"         : "lm_head",
    "unembed.W_U"     : "lm_head.weight.T",
    "unembed.b_U"     : None,
}


def build_opt_layer_map(cfg: ConfigClass):
    attn_proj_map = {"q": "q_proj", "k": "k_proj", "v": "v_proj", "o": "out_proj"}

    def opt_qkv_weight(layer, key: str, inpt: Optional[Any]=None):
        # Prepare shape changing
        their_shape = "(n_heads d_head) d_model"
        my_shape    = "n_heads d_head d_model"
        sizes = generate_sizes_dict(my_shape, cfg)

        # Get attn proj module
        attn = layer.self_attn
        attn_proj = get_attrs(attn, attn_proj_map[key])

        # Get mode
        if inpt is None:
            W = attn_proj.weight
            W = einops.rearrange(W, f"{their_shape} -> {my_shape}", **sizes)
            return W

        # Set mode
        W = einops.rearrange(inpt, f"{my_shape} -> {their_shape}", **sizes)
        update_param(attn_proj, "weight", W)

    def opt_qkv_bias(layer, key: str, inpt: Optional[Any]=None):
        # Prepare shape changing
        their_shape = "(n_heads d_head)"
        my_shape    = "n_heads d_head"
        sizes = generate_sizes_dict(my_shape, cfg)

        # Get attn proj module
        attn = layer.self_attn
        attn_proj = get_attrs(attn, attn_proj_map[key])

        if inpt is None:
            b = attn_proj.bias
            b = einops.rearrange(b, f"{their_shape} -> {my_shape}", **sizes)
            return b

        # Set mode
        b = einops.rearrange(inpt, f"{my_shape} -> {their_shape}", **sizes)
        update_param(attn_proj, "bias", b)


    opt_layer_map = {
        "attn.ln_in"           : "self_attn_layer_norm",
        "attn.ln_in.w"         : "self_attn_layer_norm.weight",
        "attn.ln_in.b"         : "self_attn_layer_norm.bias",

        "attn"          : "self_attn",
        "attn.q_proj"   : "self_attn.q_proj",
        "attn.k_proj"   : "self_attn.k_proj",
        "attn.v_proj"   : "self_attn.v_proj",

        **generate_attn_qkv_functions(opt_qkv_weight, opt_qkv_bias),

        "attn.out_proj" : "self_attn.out_proj",
        "attn.W_O"      : "self_attn.out_proj.weight",
        "attn.b_O"      : "self_attn.out_proj.bias",

        "attn.inv_out_proj" : "self_attn.inv_out_proj",
        "attn.W_O_inv"  : "self_attn.inv_out_proj.weight",
        "attn.b_O_inv"  : "self_attn.inv_out_proj.inverse_bias",

        "mlp.ln_in"           : "final_layer_norm",
        "mlp.ln_in.w"         : "final_layer_norm.weight",
        "mlp.ln_in.b"         : "final_layer_norm.bias",

        "mlp.in_proj"   : "fc1",
        "mlp.W_in"      : "fc1.weight",
        "mlp.b_in"      : "fc1.bias",

        "activation_fn" : "activation_fn",

        "mlp.out_proj"  : "fc2",
        "mlp.W_out"     : "fc2.weight",
        "mlp.b_out"     : "fc2.bias",
    }
    return opt_layer_map

# LLaMa Models
##############

llama_model_map = {
    "model"           : "model",
    "layers"          : "model.layers",
    "embed"           : "model.embed_tokens",
    "embed.W_E"       : "model.embed.weights",
    "pos_embed"       : "model.embed_positions",
    "pos_embed.W"     : "model.embed_positions.weight",
    "ln_final"        : "model.norm",
    "ln_final.w"      : "model.norm.weight",
    "unembed"         : "lm_head",
    "unembed.W_U"     : "lm_head.weight.T",
}

def build_llama_layer_map(cfg: ConfigClass):
    attn_proj_map = {"q": "q_proj", "k": "k_proj", "v": "v_proj", "o": "o_proj"}
    mlp_proj_map = {"mlp.in_proj": "up_proj", "mlp.out_proj": "down_proj", "mlp.gate_proj": "gate_proj"}

    def llama_qkv_weight(layer, key: str, inpt: Optional[Any]=None):
        # Prepare shape changing
        their_shape = "(n_heads d_head) d_model"
        my_shape    = "n_heads d_head d_model"
        sizes = generate_sizes_dict(my_shape, cfg)

        # Get attn proj module
        attn = layer.self_attn
        attn_proj = get_attrs(attn, attn_proj_map[key])

        # Get mode
        if inpt is None:
            W = attn_proj.weight
            W = einops.rearrange(W, f"{their_shape} -> {my_shape}", **sizes)
            return W

        # Set mode
        W = einops.rearrange(inpt, f"{my_shape} -> {their_shape}", **sizes)
        update_param(attn_proj, "weight", W)

    def llama_attn_bias(layer, key: str, _inpt: Optional[Any]=None):
        # Create fake bias with zeros because is easier to handle
        their_shape = "(n_heads d_head)"
        my_shape    = "n_heads d_head"
        sizes = generate_sizes_dict(my_shape, cfg)

        attn = layer.self_attn
        _proj = get_attrs(attn, attn_proj_map[key]).weight
        b = torch.zeros(
            _proj.shape[:-1], dtype=_proj.dtype, device=_proj.device
        )
        if key == "o":
            return b
        return einops.rearrange(b, f"{their_shape} -> {my_shape}", **sizes)


    def llama_mlp_bias(layer, key: str, _inpt: Optional[Any]=None):
        mlp = layer.mlp
        _proj = get_attrs(mlp, mlp_proj_map[key]).weight
        b = torch.zeros(_proj.shape[:-1], dtype=_proj.dtype, device=_proj.device)
        return b

    def get_attn_weights(attn_outputs):
        attn_out, attn_weights, past_key_value = attn_outputs
        return attn_weights

    def set_attn_weights(attn_weights, orig_output):
        return orig_output[0], attn_weights, orig_output[2]


    llama_layer_map = {
        "attn.ln_in"           : "input_layernorm",
        "attn.ln_in.w"         : "input_layernorm.weight",
        "attn.ln_in.b"         : None,

        "attn"          : "self_attn",
        "attn.q_proj"   : "self_attn.q_proj",
        "attn.k_proj"   : "self_attn.k_proj",
        "attn.v_proj"   : "self_attn.v_proj",

        **generate_attn_qkv_functions(llama_qkv_weight, llama_attn_bias),

        "attn.out_proj" : "self_attn.o_proj",
        "attn.W_O"      : "self_attn.o_proj.weight",
        "attn.b_O"      : lambda layer, _inpt=None: llama_attn_bias(layer, "o", _inpt),

        "attn.inv_out_proj" : "self_attn.inv_out_proj",
        "attn.W_O_inv"  : "self_attn.inv_out_proj.weight",
        "attn.b_O_inv"  : "self_attn.inv_out_proj.bias",

        "mlp.ln_in"           : "post_attention_layernorm",
        "mlp.ln_in.w"         : "post_attention_layernorm.weight",
        "mlp.ln_in.b"         : None,

        "mlp"           : "mlp",
        "mlp.in_proj"   : "mlp.up_proj",
        "mlp.gate_proj" : "mlp.gate_proj",
        "mlp.W_in"      : "mlp.up_proj.weight",
        "mlp.W_gate"    : "mlp.gate_proj.weight",
        "mlp.b_in"      : lambda layer, _inpt=None: llama_mlp_bias(layer, "mlp.in_proj", _inpt),
        "mlp.b_gate"    : lambda layer, _inpt=None: llama_mlp_bias(layer, "mlp.out_proj", _inpt),

        "activation_fn" : "mlp.act_fn",

        "mlp.out_proj"  : "mlp.down_proj",
        "mlp.W_out"     : "mlp.down_proj.weight",
        "mlp.b_out"     : lambda layer, _inpt=None: llama_mlp_bias(layer, "mlp.gate_proj", _inpt),
    }
    return llama_layer_map

# Mistral Model
###############

mistral_model_map = {
    "model"           : "model",
    "layers"          : "model.layers",
    "embed"           : "model.embed_tokens",
    "embed.W_E"       : "model.embed.weights",
    "pos_embed"       : "model.embed_positions",
    "pos_embed.W"     : "model.embed_positions.weight",
    "ln_final"        : "model.norm",
    "ln_final.w"      : "model.norm.weight",
    "unembed"         : "model.lm_head",
    "unembed.W_U"     : "model.lm_head.weight.T",
    "unembed.b_U"     : None,
}

def build_mistral_layer_map(cfg: ConfigClass):
    attn_proj_map = {"q": "q_proj", "k": "k_proj", "v": "v_proj", "o": "o_proj"}
    mlp_proj_map = {"mlp.in_proj": "up_proj", "mlp.out_proj": "down_proj", "mlp.gate_proj": "gate_proj"}

    def mistral_qkv_weight(layer, key: str, inpt: Optional[Any]=None):
        # Prepare shape changing
        if key in ['k', 'v']:  # Adjust for Mistral's unique n_key_value_heads
            their_shape = "(n_key_value_heads d_head) d_model"
        else:
            their_shape = "(n_heads d_head) d_model"
        my_shape = "n_heads d_head d_model"
        sizes = generate_sizes_dict(my_shape, cfg)

        # Get attn proj module
        attn = layer.self_attn
        attn_proj = get_attrs(attn, attn_proj_map[key])

        # Get mode
        if inpt is None:
            W = attn_proj.weight
            W = einops.rearrange(W, f"{their_shape} -> {my_shape}", **sizes)
            return W

        # Set mode
        W = einops.rearrange(inpt, f"{my_shape} -> {their_shape}", **sizes)
        update_param(attn_proj, "weight", W)

    def mistral_attn_bias(layer, key: str, _inpt: Optional[Any]=None):
        # Create fake bias with zeros because is easier to handle
        their_shape = "(n_heads d_head)"
        my_shape = "n_heads d_head"
        sizes = generate_sizes_dict(my_shape, cfg)

        attn = layer.self_attn
        _proj = get_attrs(attn, attn_proj_map[key]).weight
        b = torch.zeros(_proj.shape[:-1], dtype=_proj.dtype, device=_proj.device)
        if key == "o":
            return b
        return einops.rearrange(b, f"{their_shape} -> {my_shape}", **sizes)

    def mistral_mlp_bias(layer, key: str, _inpt: Optional[Any]=None):
        mlp = layer.mlp
        _proj = get_attrs(mlp, mlp_proj_map[key]).weight
        b = torch.zeros(_proj.shape[:-1], dtype=_proj.dtype, device=_proj.device)
        return b

    mistral_layer_map = {
        "attn.ln_in"           : "input_layernorm",
        "attn.ln_in.w"         : "input_layernorm.weight",
        "attn.ln_in.b"         : None,

        "attn"          : "self_attn",
        "attn.q_proj"   : "self_attn.q_proj",
        "attn.k_proj"   : "self_attn.k_proj",
        "attn.v_proj"   : "self_attn.v_proj",

        **generate_attn_qkv_functions(mistral_qkv_weight, mistral_attn_bias),

        "attn.inv_out_proj" : "self_attn.inv_out_proj",
        "attn.out_proj" : "self_attn.o_proj",
        "attn.W_O"      : "self_attn.o_proj.weight",
        "attn.b_O"      : lambda layer, _inpt=None: mistral_attn_bias(layer, "o", _inpt),

        "mlp.ln_in"           : "post_attention_layernorm",
        "mlp.ln_in.w"         : "post_attention_layernorm.weight",
        "mlp.ln_in.b"         : None,

        "mlp"           : "mlp",
        "mlp.in_proj"   : "mlp.up_proj",
        "mlp.gate_proj" : "mlp.gate_proj",
        "mlp.W_in"      : "mlp.up_proj.weight",
        "mlp.W_gate"    : "mlp.gate_proj.weight",
        "mlp.b_in"      : lambda layer, _inpt=None: mistral_mlp_bias(layer, "mlp.in_proj", _inpt),
        "mlp.b_gate"    : lambda layer, _inpt=None: mistral_mlp_bias(layer, "mlp.out_proj", _inpt),

        "activation_fn" : "mlp.act_fn",

        "mlp.out_proj"  : "mlp.down_proj",
        "mlp.W_out"     : "mlp.down_proj.weight",
        "mlp.b_out"     : lambda layer, _inpt=None: mistral_mlp_bias(layer, "mlp.gate_proj", _inpt),

    }

    return mistral_layer_map


# GEMMA
#######

gemma_model_map = {
"model" : "model",
"layers" : "model.layers",
"embed" : "model.embed_tokens",
"embed.W_E" : "model.embed.weights",
"pos_embed" : "model.embed_positions",
"pos_embed.W" : "model.embed_positions.weight",
"ln_final" : "model.norm",
"ln_final.w" : "model.norm.weight",
"unembed"     : "lm_head",
"unembed.W_U" : "lm_head.weight.T",
"unembed.b_U" : None,
}

def build_gemma_layer_map(cfg: ConfigClass):
    attn_proj_map = {"q": "q_proj", "k": "k_proj", "v": "v_proj", "o": "o_proj"}
    mlp_proj_map = {"mlp.in_proj": "up_proj", "mlp.out_proj": "down_proj", "mlp.gate_proj": "gate_proj"}

    def gemma_qkv_weight(layer, key: str, inpt: Optional[Any]=None):
        # Prepare shape changing
        their_shape = "(n_heads d_head) d_model"
        my_shape    = "n_heads d_head d_model"
        sizes = generate_sizes_dict(my_shape, cfg)

        # Get attn proj module
        attn = layer.self_attn
        attn_proj = get_attrs(attn, attn_proj_map[key])

        # Get mode
        if inpt is None:
            W = attn_proj.weight
            W = einops.rearrange(W, f"{their_shape} -> {my_shape}", **sizes)
            return W

        # Set mode
        W = einops.rearrange(inpt, f"{my_shape} -> {their_shape}", **sizes)
        update_param(attn_proj, "weight", W)

    def gemma_attn_bias(layer, key: str, _inpt: Optional[Any]=None):
        # Create fake bias with zeros because is easier to handle
        their_shape = "(n_heads d_head)"
        my_shape    = "n_heads d_head"
        sizes = generate_sizes_dict(my_shape, cfg)

        attn = layer.self_attn
        _proj = get_attrs(attn, attn_proj_map[key]).weight
        b = torch.zeros(
            _proj.shape[:-1], dtype=_proj.dtype, device=_proj.device
        )
        if key == "o":
            return b
        return einops.rearrange(b, f"{their_shape} -> {my_shape}", **sizes)


    def gemma_mlp_bias(layer, key: str, _inpt: Optional[Any]=None):
        mlp = layer.mlp
        _proj = get_attrs(mlp, mlp_proj_map[key]).weight
        b = torch.zeros(_proj.shape[:-1], dtype=_proj.dtype, device=_proj.device)
        return b

    gemma_layer_map = {
        "attn.ln_in"           : "input_layernorm",
        "attn.ln_in.w"         : "input_layernorm.weight",
        "attn.ln_in.b"         : None,

        "attn"          : "self_attn",
        "attn.q_proj"   : "self_attn.q_proj",
        "attn.k_proj"   : "self_attn.k_proj",
        "attn.v_proj"   : "self_attn.v_proj",

        **generate_attn_qkv_functions(gemma_qkv_weight, gemma_attn_bias),

        "attn.out_proj" : "self_attn.o_proj",
        "attn.W_O"      : "self_attn.o_proj.weight",
        "attn.b_O"      : lambda layer, _inpt=None: gemma_attn_bias(layer, "o", _inpt),

        "attn.inv_out_proj" : "self_attn.inv_out_proj",
        "attn.W_O_inv"  : "self_attn.inv_out_proj.weight",
        "attn.b_O_inv"  : "self_attn.inv_out_proj.bias",

        "mlp.ln_in"           : "post_attention_layernorm",
        "mlp.ln_in.w"         : "post_attention_layernorm.weight",
        "mlp.ln_in.b"         : None,

        "mlp"           : "mlp",
        "mlp.in_proj"   : "mlp.up_proj",
        "mlp.gate_proj" : "mlp.gate_proj",
        "mlp.W_in"      : "mlp.up_proj.weight",
        "mlp.W_gate"    : "mlp.gate_proj.weight",
        "mlp.b_in"      : lambda layer, _inpt=None: gemma_mlp_bias(layer, "mlp.in_proj", _inpt),
        "mlp.b_gate"    : lambda layer, _inpt=None: gemma_mlp_bias(layer, "mlp.gate_proj", _inpt),

        "activation_fn" : "mlp.act_fn",

        "mlp.out_proj"  : "mlp.down_proj",
        "mlp.W_out"     : "mlp.down_proj.weight",
        "mlp.b_out"     : lambda layer, _inpt=None: gemma_mlp_bias(layer, "mlp.out_proj", _inpt),
    }
    return gemma_layer_map

# Gemma 2
#########

gemma2_model_map = {
    "model"           : "model",
    "layers"          : "model.layers",
    "embed"           : "model.embed_tokens",
    "embed.W_E"       : "model.embed_tokens.weight",
    "ln_final"        : "model.norm",
    "ln_final.w"      : "model.norm.weight",
    "unembed"         : "lm_head",
    "unembed.W_U"     : "lm_head.weight.T",
    "unembed.b_U"     : None,
}

def build_gemma2_layer_map(cfg: ConfigClass):
    attn_proj_map = {"q": "q_proj", "k": "k_proj", "v": "v_proj", "o": "o_proj"}
    mlp_proj_map = {"mlp.gate_proj": "gate_proj", "mlp.up_proj": "up_proj", "mlp.down_proj": "down_proj"}

    def gemma2_qkv_weight(layer, key: str, inpt: Optional[Any]=None):
        # Prepare shape changing
        their_shape = "(n_heads d_head) d_model"
        my_shape    = "n_heads d_head d_model"
        sizes = generate_sizes_dict(my_shape, cfg)

        # Get attn proj module
        attn = layer.self_attn
        attn_proj = get_attrs(attn, attn_proj_map[key])

        # Get mode
        if inpt is None:
            W = attn_proj.weight
            W = einops.rearrange(W, f"{their_shape} -> {my_shape}", **sizes)
            return W

        # Set mode
        W = einops.rearrange(inpt, f"{my_shape} -> {their_shape}", **sizes)
        update_param(attn_proj, "weight", W)

    def gemma2_attn_bias(layer, key: str, _inpt: Optional[Any]=None):
        # Gemma2 doesn't use biases in attention
        return None

    def gemma2_mlp_weight(layer, key: str, inpt: Optional[Any]=None):
        mlp = layer.mlp
        proj = get_attrs(mlp, mlp_proj_map[key])

        if inpt is None:
            return proj.weight
        update_param(proj, "weight", inpt)

    def gemma2_mlp_bias(layer, key: str, _inpt: Optional[Any]=None):
        # Gemma2 doesn't use biases in MLP
        return None

    gemma2_layer_map = {
        "attn.ln_in"    : "input_layernorm",
        "attn.ln_in.w"  : "input_layernorm.weight",
        "attn.ln_in.b"  : None,

        "attn"          : "self_attn",
        "attn.q_proj"   : "self_attn.q_proj",
        "attn.k_proj"   : "self_attn.k_proj",
        "attn.v_proj"   : "self_attn.v_proj",

        **generate_attn_qkv_functions(gemma2_qkv_weight, gemma2_attn_bias),

        "attn.out_proj" : "self_attn.o_proj",
        "attn.W_O"      : "self_attn.o_proj.weight",
        "attn.b_O"      : None,

        "attn.ln_out"   : "post_attention_layernorm",
        "attn.ln_out.w" : "post_attention_layernorm.weight",
        "attn.ln_out.b" : None,

        "mlp"           : "mlp",
        "mlp.gate_proj" : "mlp.gate_proj",
        "mlp.up_proj"   : "mlp.up_proj",
        "mlp.out_proj" : "mlp.down_proj",
        "mlp.W_gate"    : lambda layer, inpt=None: gemma2_mlp_weight(layer, "mlp.gate_proj", inpt),
        "mlp.W_up"      : lambda layer, inpt=None: gemma2_mlp_weight(layer, "mlp.up_proj", inpt),
        "mlp.W_out"    : lambda layer, inpt=None: gemma2_mlp_weight(layer, "mlp.down_proj", inpt),
        "mlp.b_gate"    : lambda layer, _inpt=None: gemma2_mlp_bias(layer, "mlp.gate_proj", _inpt),
        "mlp.b_up"      : lambda layer, _inpt=None: gemma2_mlp_bias(layer, "mlp.up_proj", _inpt),
        "mlp.b_out"    : lambda layer, _inpt=None: gemma2_mlp_bias(layer, "mlp.down_proj", _inpt),

        "activation_fn" : "mlp.act_fn",

        "mlp.ln_in"     : "pre_feedforward_layernorm",
        "mlp.ln_in.w"   : "pre_feedforward_layernorm.weight",
        "mlp.ln_in.b"   : None,

        "mlp.ln_out"    : "post_feedforward_layernorm",
        "mlp.ln_out.w"  : "post_feedforward_layernorm.weight",
        "mlp.ln_out.b"  : None,
    }

    return gemma2_layer_map

# Gemma 3
#########

gemma3_textonly_model_map = {
    "model"           : "model",
    "layers"          : "model.layers",
    "embed"           : "model.embed_tokens",
    "embed.W_E"       : "model.embed_tokens.weight",
    "ln_final"        : "model.norm",
    "ln_final.w"      : "model.norm.weight",
    "unembed"         : "lm_head",
}

gemma3_model_map = {
    "model"           : "language_model.model", # In Gemma3ForCausalLM, it's self.model = Gemma3TextModel(config)
    "layers"          : "language_model.model.layers",
    "embed"           : "language_model.model.embed_tokens",
    "embed.W_E"       : "language_model.model.embed_tokens.weight", # Uses Gemma3TextScaledWordEmbedding
    # Positional embedding handled by rotary_emb / rotary_emb_local inside the model/layers
    "ln_final"        : "language_model.model.norm",
    "ln_final.w"      : "language_model.model.norm.weight",
    "unembed"         : "language_model.lm_head", # In Gemma3ForCausalLM
    "unembed.W_U"     : "language_model.lm_head.weight.T", # lm_head is Linear with bias=False
    "unembed.b_U"     : None,
}

def build_gemma3_layer_map(cfg: ConfigClass):
    attn_proj_map = {"q": "q_proj", "k": "k_proj", "v": "v_proj", "o": "o_proj"}
    mlp_proj_map = {"mlp.gate_proj": "gate_proj", "mlp.up_proj": "up_proj", "mlp.down_proj": "down_proj"}

    def gemma3_qkv_weight(layer, key: str, inpt: Optional[Any]=None):
        # Prepare shape changing - handle MQA/GQA like Mistral/Gemma2
        is_kv = key in ['k', 'v']
        num_heads = cfg.n_key_value_heads if is_kv else cfg.n_heads
        their_shape = f"({num_heads} d_head) d_model"
        my_shape = f"{num_heads} d_head d_model"
        sizes = generate_sizes_dict(my_shape, cfg)
        sizes["n_heads"] = num_heads # Override n_heads for K/V GQA case

        # Get attn proj module
        attn = layer.self_attn
        attn_proj = get_attrs(attn, attn_proj_map[key])

        # Get mode
        if inpt is None:
            W = attn_proj.weight
            W = einops.rearrange(W, f"{their_shape} -> {my_shape}", **sizes)
            return W

        # Set mode
        W = einops.rearrange(inpt, f"{my_shape} -> {their_shape}", **sizes)
        update_param(attn_proj, "weight", W)

    def gemma3_attn_bias(layer, key: str, inpt: Optional[Any]=None):
        # Check if attention bias is configured (assuming it might be)
        # default in Gemma3TextConfig seems to be False for attention_bias
        # For simplicity, let's assume bias=False based on the provided code's defaults.
        # If attention_bias=True needs to be supported, check layer.self_attn.[qkv]_proj.bias
        # attn = layer.self_attn
        # attn_proj = get_attrs(attn, attn_proj_map[key])
        # if attn_proj.bias is None:
        #     return None
        #
        # # Prepare shape changing like gemma3_qkv_weight if bias exists
        # is_kv = key in ['k', 'v']
        # num_heads = cfg.n_key_value_heads if is_kv else cfg.n_heads
        # their_shape = f"({num_heads} d_head)"
        # my_shape = f"{num_heads} d_head"
        # sizes = generate_sizes_dict(my_shape, cfg)
        # sizes["n_heads"] = num_heads
        #
        # if inpt is None:
        #     b = attn_proj.bias
        #     b = einops.rearrange(b, f"{their_shape} -> {my_shape}", **sizes)
        #     return b
        #
        # # Set mode
        # b = einops.rearrange(inpt, f"{my_shape} -> {their_shape}", **sizes)
        # update_param(attn_proj, "bias", b)
        return None # Assuming bias=False based on provided code examples

    def gemma3_mlp_weight(layer, key: str, inpt: Optional[Any]=None):
        mlp = layer.mlp
        proj = get_attrs(mlp, mlp_proj_map[key])

        if inpt is None:
            return proj.weight
        update_param(proj, "weight", inpt)

    def gemma3_mlp_bias(layer, key: str, _inpt: Optional[Any]=None):
        # Gemma3 MLP uses bias=False
        return None

    # Maps internal names to Gemma3 HF names
    gemma3_layer_map = {
        # Input Norm (Before Attention)
        "attn.ln_in"    : "input_layernorm",
        "attn.ln_in.w"  : "input_layernorm.weight",
        "attn.ln_in.b"  : None, # RMSNorm

        # Attention Projections
        "attn"          : "self_attn",
        "attn.q_proj"   : "self_attn.q_proj",
        "attn.k_proj"   : "self_attn.k_proj",
        "attn.v_proj"   : "self_attn.v_proj",

        **generate_attn_qkv_functions(gemma3_qkv_weight, gemma3_attn_bias),

        # Attention Output Projection
        "attn.out_proj" : "self_attn.o_proj",
        "attn.W_O"      : "self_attn.o_proj.weight",
        "attn.b_O"      : lambda layer, inpt=None: gemma3_attn_bias(layer, "o", inpt), # Use bias func

        # Post-Attention Norm
        "attn.ln_out"   : "post_attention_layernorm",
        "attn.ln_out.w" : "post_attention_layernorm.weight",
        "attn.ln_out.b" : None, # RMSNorm

        # Pre-MLP Norm
        "mlp.ln_in"     : "pre_feedforward_layernorm",
        "mlp.ln_in.w"   : "pre_feedforward_layernorm.weight",
        "mlp.ln_in.b"   : None, # RMSNorm

        # MLP Layers
        "mlp"           : "mlp",
        "mlp.gate_proj" : "mlp.gate_proj",
        "mlp.up_proj"   : "mlp.up_proj", # Maps to W_in conceptually
        "mlp.out_proj"  : "mlp.down_proj", # Maps to W_out conceptually
        "mlp.W_gate"    : lambda layer, inpt=None: gemma3_mlp_weight(layer, "mlp.gate_proj", inpt),
        "mlp.W_in"      : lambda layer, inpt=None: gemma3_mlp_weight(layer, "mlp.up_proj", inpt), # W_in maps to up_proj
        "mlp.W_out"     : lambda layer, inpt=None: gemma3_mlp_weight(layer, "mlp.down_proj", inpt),
        "mlp.b_gate"    : lambda layer, _inpt=None: gemma3_mlp_bias(layer, "mlp.gate_proj", _inpt),
        "mlp.b_in"      : lambda layer, _inpt=None: gemma3_mlp_bias(layer, "mlp.up_proj", _inpt), # b_in maps to up_proj
        "mlp.b_out"     : lambda layer, _inpt=None: gemma3_mlp_bias(layer, "mlp.down_proj", _inpt),

        # Activation Function
        "activation_fn" : "mlp.act_fn",

        # Post-MLP Norm
        "mlp.ln_out"    : "post_feedforward_layernorm",
        "mlp.ln_out.w"  : "post_feedforward_layernorm.weight",
        "mlp.ln_out.b"  : None, # RMSNorm
    }

    return gemma3_layer_map

# PHI 1 and 2 models
####################

phi_model_map = {
    "model": "model",
    "layers": "model.layers",
    "embed": "model.embed_tokens",
    "embed.W_E": "model.embed_tokens.weight",
    "ln_final": "model.final_layernorm",
    "ln_final.w": "model.final_layernorm.weight",
    "ln_final.b": "model.final_layernorm.bias",
    "unembed"    : "lm_head",
    "unembed.W_U": "lm_head.weight.T",
    "unembed.b_U": "lm_head.bias",
}

def build_phi_layer_map(cfg: ConfigClass):
    attn_proj_map = {"q": "q_proj", "k": "k_proj", "v": "v_proj", "o": "dense"}
    mlp_proj_map = {"mlp.W_in": "fc1", "mlp.W_out": "fc2"}

    def phi_qkv_weight(layer, key: str, inpt: Optional[Any]=None):
        # Prepare shape changing
        their_shape = "(n_head d_head) d_model"
        my_shape = "n_head d_model d_head"
        sizes = generate_sizes_dict(my_shape, cfg)

        # Get attn proj module
        attn = layer.self_attn
        attn_proj = get_attrs(attn, attn_proj_map[key])

        # Get mode
        if inpt is None:
            W = attn_proj.weight
            W = einops.rearrange(W, f"{their_shape} -> {my_shape}", **sizes)
            return W
        # Set mode
        W = einops.rearrange(inpt, f"{my_shape} -> {their_shape}", **sizes)
        update_param(attn_proj, "weight", W)

    def phi_attn_bias(layer, key: str, inpt: Optional[Any]=None):
        # Prepare shape changing
        their_shape = "(n_head d_head)"
        my_shape = "n_head d_head"
        sizes = generate_sizes_dict(my_shape, cfg)

        # Get attn proj module
        attn = layer.self_attn
        attn_proj = get_attrs(attn, attn_proj_map[key])

        # Get mode
        if inpt is None:
            b = attn_proj.bias
            b = einops.rearrange(b, f"{their_shape} -> {my_shape}", **sizes)
            return b
        # Set mode
        b = einops.rearrange(inpt, f"{my_shape} -> {their_shape}", **sizes)
        update_param(attn_proj, "bias", b)

    def phi_mlp_bias(layer, key: str, inpt: Optional[Any]=None):
        mlp = layer.mlp
        _proj = get_attrs(mlp, mlp_proj_map[key]).weight
        b = torch.zeros(_proj.shape[:-1], dtype=_proj.dtype, device=_proj.device)
        return b

    phi_layer_map = {
        "attn.ln_in":   "input_layernorm",
        "attn.ln_in.w": "input_layernorm.weight",
        "attn.ln_in.b": "input_layernorm.bias",
        "attn": "self_attn",
        **generate_attn_qkv_functions(phi_qkv_weight, phi_attn_bias),
        "attn.W_O": "self_attn.dense.weight",
        "attn.b_O": "self_attn.dense.bias",
        "mlp.ln_in": "input_layernorm", # parallel attn mlp
        "mlp.ln_in.w": "input_layernorm.weight",
        "mlp.ln_in.b": "input_layernorm.bias",
        "mlp": "mlp",
        "activation_fn" : "mlp.act_fn",
        "mlp.W_in"      : "mlp.fc1.weight",
        "mlp.b_in"      : lambda layer, _inpt=None: phi_mlp_bias(layer, "mlp.fc1", _inpt),
        "mlp.W_gate"    : "mlp.fc2.weight",
        "mlp.b_gate"    : lambda layer, _inpt=None: phi_mlp_bias(layer, "mlp.fc2", _inpt),
    }

    return phi_layer_map

# PHI-3
#######

phi3_model_map = {
    "model": "model",
    "layers": "model.layers",
    "embed": "model.embed_tokens",
    "embed.W_E": "model.embed_tokens.weight",
    "ln_final": "model.norm",
    "ln_final.w": "model.norm.weight",
    "ln_final.b": "model.norm.bias",
    "unembed"    : "lm_head",
    "unembed.W_U": "lm_head.weight.T",
    "unembed.b_U": "lm_head.bias",
}

def build_phi3_layer_map(cfg: ConfigClass):
    attn_proj_map = {"q": "q_proj", "k": "k_proj", "v": "v_proj", "o": "dense"}

    def phi3_qkv_weight(layer, key: str, inpt: Optional[Any]=None):
        # Prepare shape changing
        their_shape = "(n_head d_head) d_model"
        my_shape = "n_head d_model d_head"
        sizes = generate_sizes_dict(my_shape, cfg)

        # Attention qkv proj is combined in phi:
        # op_size = self.num_heads * self.head_dim + 2 * (self.num_key_value_heads * self.head_dim)
        # self.o_proj = nn.Linear(self.num_heads * self.head_dim, self.hidden_size, bias=False)
        # self.qkv_proj = nn.Linear(self.hidden_size, op_size, bias=False)

        attn = layer.self_attn
        attn_proj = get_attrs(attn, attn_proj_map[key])

        # Get mode
        if inpt is None:
            W = attn_proj.weight
            W = einops.rearrange(W, f"{their_shape} -> {my_shape}", **sizes)
            return W
        # Set mode
        W = einops.rearrange(inpt, f"{my_shape} -> {their_shape}", **sizes)
        update_param(attn_proj, "weight", W)

    def phi3_attn_bias(layer, key: str, inpt: Optional[Any]=None):
        # Get attn proj module
        attn = layer.self_attn
        attn_proj = get_attrs(attn, attn_proj_map[key])

        # Get mode
        if inpt is None:
            b = attn_proj.bias
            return b
        # Set mode
        update_param(attn_proj, "bias", inpt)

    def phi3_mlp_bias(layer, key: str, _inpt: Optional[Any]=None):
        mlp = layer.mlp
        if key == "mlp.b_in":
            _proj = mlp.gate_up_proj.weight
        elif key == "mlp.b_out":
            _proj = mlp.down_proj.weight
        else:
            raise ValueError(f"Unknown MLP bias key {key}")

        b = torch.zeros(_proj.shape[1] // 2 if key == "mlp.b_in" else _proj.shape[0], dtype=_proj.dtype)
        return b

    phi3_layer_map = {
        "attn.ln_in": "input_layernorm",
        "attn.ln_in.w": "input_layernorm.weight",
        "attn.ln_in.b": "input_layernorm.bias",

        "attn": "self_attn",
        **generate_attn_qkv_functions(phi3_qkv_weight, phi3_attn_bias),
        "attn.out_proj": "self_attn.o_proj",
        "attn.W_O": "self_attn.o_proj.weight",
        "attn.b_O": "self_attn.o_proj.bias",

        "mlp.ln_in"  : "post_attention_layernorm",
        "mlp.ln_in.w": "post_attention_layernorm.weight",
        "mlp.ln_in.b": None,

        "mlp": "mlp",
        #"mlp.in_proj" : "mlp.gate_up_proj", # this combined up and gate proj matrices
        "mlp.out_proj": "mlp.down_proj",
        #"mlp.W_in": "mlp.gate_up_proj.weight",
        #"mlp.b_in": lambda layer, _inpt=None: phi3_mlp_bias(layer, "mlp.b_in", _inpt),
        "mlp.W_out": "mlp.down_proj.weight",
        "mlp.b_out": lambda layer, _inpt=None: phi3_mlp_bias(layer, "mlp.b_out", _inpt),
        "activation_fn": "mlp.activation_fn",
    }

    return phi3_layer_map


# GPT NEO X and Pythia Models
#############################

gpt_neox_model_map = {
    "model"           : "base_model",
    "layers"          : "base_model.layers",
    "embed"           : "base_model.embed_in",
    "embed.W_E"       : "base_model.embed_in.weight",
    "pos_embed.W_pos" : "base_model.embed_pos.weight",
    "ln_final"        : "base_model.final_layer_norm",
    "ln_final.w"      : "base_model.final_layer_norm.weight",
    "ln_final.b"      : "base_model.final_layer_norm.bias",
    "unembed"         : "base_model.embed_out",
    "unembed.W_U"     : "base_model.embed_out.weight",
    "unembed.b_U"     : "base_model.embed_out.bias",
}

def build_gpt_neox_layer_map(cfg: ConfigClass):
    def gpt_neox_qkv_weight(layer, key: str, inpt: Optional[Any]=None):
        # Prepare shape changing
        their_shape = "(n_heads qkv d_head) d_model"
        my_shape    = "qkv n_heads d_head d_model"
        sizes = generate_sizes_dict(my_shape, cfg)
        qkv_map = {"q": 0, "k": 1, "v": 2}
        index = qkv_map[key]

        # Get the head weights
        qkv_heads = layer.attention.query_key_value
        W = qkv_heads.weight
        W = einops.rearrange(W, f"{their_shape} -> {my_shape}", **sizes)

        # Get mode
        if inpt is None:
            return W[index]

        # Set mode
        W[index] = inpt
        W = einops.rearrange(W, f"{my_shape} -> {their_shape}", **sizes)
        update_param(qkv_heads, "weight", W)

    def gpt_neox_qkv_bias(layer, key: str, inpt: Optional[Any]=None):
        # Prepare shape changing
        their_shape = "(n_heads qkv d_head)"
        my_shape    = "qkv n_heads d_head"
        sizes = generate_sizes_dict(my_shape, cfg)
        qkv_map = {"q": 0, "k": 1, "v": 2}
        index = qkv_map[key]

        # Get the head biases
        qkv_head = layer.attention.query_key_value
        qkv_bias = qkv_head.bias
        qkv_bias = einops.rearrange(qkv_bias, f"{their_shape} -> {my_shape}", **sizes)

        # Get mode
        if inpt is None:
            return qkv_bias[index]

        # Set mode
        qkv_bias[index] = inpt
        qkv_bias = einops.rearrange(qkv_bias, f"{my_shape} -> {their_shape}", **sizes)
        update_param(qkv_head, "bias", qkv_bias)

    gpt_neox_layer_map = {
        "attn.ln_in"       : "input_layernorm",
        "attn.ln_in.w"     : "input_layernorm.weight",
        "attn.ln_in.b"     : "input_layernorm.bias",

        "attn"      : "attention",
        "attn.q_proj"   : None,
        "attn.k_proj"   : None,
        "attn.v_proj"   : None,

        **generate_attn_qkv_functions(gpt_neox_qkv_weight, gpt_neox_qkv_bias),

        "attn.out_proj" : "attention.dense",
        "attn.W_O"      : "attention.dense.weight",
        "attn.b_O"      : "attention.dense.bias",

        "attn.inv_out_proj" : "attention.inv_out_proj",
        "attn.W_O_inv"      : "attention.inv_out_proj.weight",
        "attn.b_O_inv"      : "attention.inv_out_proj.inverse_bias",

        "mlp.ln_in"       : "post_attention_layernorm",
        "mlp.ln_in.w"     : "post_attention_layernorm.weight",
        "mlp.ln_in.b"     : "post_attention_layernorm.bias",

        "mlp"         : "mlp",
        "mlp.in_proj" : "mlp.dense_h_to_4h",
        "mlp.out_proj": "mlp.dense_4h_to_h",
        "mlp.W_in"  : "mlp.dense_h_to_4h.weight",
        "mlp.W_out" : "mlp.dense_4h_to_h.weight",
        "mlp.b_in"  : "mlp.dense_h_to_4h.bias",
        "mlp.b_out" : "mlp.dense_4h_to_h.bias",
        "activation_fn" : "mlp.act",
    }
    return gpt_neox_layer_map

# GPT2 Models
#############

gpt2_model_map = {
    "model"           : "transformer",
    "layers"          : "transformer.h",
    "embed"           : "transformer.wte",
    "embed.W_E"       : "transformer.wte.weight",
    "pos_embed.W_pos" : "transformer.wpe",
    "ln_final"        : "transformer.ln_f",
    "ln_final.w"      : "transformer.ln_f.weight",
    "ln_final.b"      : "transformer.ln_f.bias",
    "unembed"         : "lm_head",
    "unembed.W_U"     : "lm_head.weight.T",
    "unembed.b_U"     : None,
}

def build_gpt2_layer_map(cfg: ConfigClass):
    def gpt2_qkv_weight(layer, key: str, inpt: Optional[Any]=None):
        their_shape = "d_model (qkv n_heads d_head)"
        my_shape    = "qkv n_heads d_head d_model"
        sizes = generate_sizes_dict(my_shape, cfg)
        qkv_map = {"q": 0, "k": 1, "v": 2}
        index = qkv_map[key]

        # Get the head weights
        qkv_heads = layer.attn.c_attn
        W = qkv_heads.weight
        W = einops.rearrange(W, f"{their_shape} -> {my_shape}", **sizes)

        # Get mode
        if inpt is None:
            return W[index]

        # Set mode
        W[index] = inpt
        W = einops.rearrange(W, f"{my_shape} -> {their_shape}", **sizes)
        update_param(qkv_heads, "weight", W)

    def gpt2_qkv_bias(layer, key: str, inpt: Optional[Any]=None):
        their_shape = "(qkv n_heads d_head)"
        my_shape    = "qkv n_heads d_head"
        sizes = generate_sizes_dict(my_shape, cfg)
        qkv_map = {"q": 0, "k": 1, "v": 2}
        index = qkv_map[key]

        # Get the head biases
        qkv_heads = layer.attn.c_attn
        qkv_bias = qkv_heads.bias
        qkv_bias = einops.rearrange(qkv_bias, f"{their_shape} -> {my_shape}", **sizes)

        # Get mode
        if inpt is None:
            return qkv_bias[index]

        # Set mode
        qkv_bias[index] = inpt
        qkv_bias = einops.rearrange(qkv_bias, f"{my_shape} -> {their_shape}", **sizes)
        update_param(qkv_heads, "bias", qkv_bias)

    # GPT2 uses Conv1D instead of Linear, so we must get the transpose
    def conv1d_weight(module, inpt=None):
        if inpt is None:
            return module.weight.T
        params = module.state_dict()
        params["weight"] = inpt.T
        module.load_state_dict(params)

    def gpt2_out_weight(layer, inpt=None):
        return conv1d_weight(layer.attn.c_proj, inpt)

    def gpt2_mlp_in_weight(layer, inpt=None):
        return conv1d_weight(layer.mlp.c_fc, inpt)

    def gpt2_mlp_out_weight(layer, inpt=None):
        return conv1d_weight(layer.mlp.c_proj, inpt)

    def get_attn_weight(attn_outputs):
            # outputs # a, present, (attentions)
            # TODO: make sure use_cache is true!
            attn_out, key_value_cache, attn_weights = attn_outputs
            return attn_weights


    gpt2_layer_map = {
        "attn.ln_out"       : "ln_1",
        "attn.ln_out.w"     : "ln_1.weight",
        "attn.ln_out.b"     : "ln_1.bias",
        "attn"      : "attn",
        "attn.q_proj"   : None,
        "attn.k_proj"   : None,
        "attn.v_proj"   : None,
        **generate_attn_qkv_functions(gpt2_qkv_weight, gpt2_qkv_bias),
        "attn.out_proj" : "attn.c_proj",
        "attn.W_O"      : lambda layer, inpt=None: gpt2_out_weight(layer, inpt),
        "attn.b_O"      : "attn.c_proj.bias",
        "attn.inv_out_proj" : "attn.inv_out_proj",
        "attn.W_O_inv"      : "attn.inv_out_proj.weight",
        "attn.b_O_inv"      : "attn.inv_out_proj.inverse_bias",
        "mlp.ln_out"       : "ln_2",
        "mlp.ln_out.w"     : "ln_2.weight",
        "mlp.ln_out.b"     : "ln_2.bias",
        "mlp"         : "mlp",
        "mlp.in_proj" : "mlp.c_fc",
        "mlp.out_proj": "mlp.c_proj",
        "mlp.W_in"  : lambda layer, inpt=None: gpt2_mlp_in_weight(layer, inpt),
        "mlp.W_out" : lambda layer, inpt=None: gpt2_mlp_out_weight(layer, inpt),
        "mlp.b_in"  : "mlp.c_fc.bias",
        "mlp.b_out" : "mlp.c_proj.bias",
        "activation_fn" : "mlp.act",
    }
    return gpt2_layer_map

#####################################################################################
# "Masked LM" Models
#####################################################################################

# Roberta Model Map
###################

roberta_model_map = {
    "model"           : "roberta",
    "layers"          : "roberta.encoder.layer",
    "embed"           : "roberta.embeddings",
    "embed.W_E"       : "roberta.embeddings.word_embeddings.weight",
    "pos_embed.W_pos" : "roberta.embeddings.position_embeddings",
    "ln_final"        : "lm_head.layer_norm",
    "ln_final.w"      : "lm_head.layer_norm.weight",
    "ln_final.b"      : "lm_head.layer_norm.bias",
    "unembed"         : "lm_head",
    "unembed.W_U"     : "lm_head.dense.weight",
    "unembed.b_U"     : None,
}

def build_roberta_layer_map(cfg: ConfigClass):
    attn_proj_map = {
        "q": "self.query",
        "k": "self.key",
        "v": "self.value",
        "o": "output.dense"
    }

    def roberta_qkv_weight(layer, key: str, inpt: Optional[Any]=None):
        # Prepare shape changing
        their_shape = "(n_heads d_head) d_model"
        my_shape    = "n_heads d_head d_model"
        sizes = generate_sizes_dict(my_shape, cfg)

        # Get attn proj module
        attn = layer.attention
        attn_proj = get_attrs(attn, attn_proj_map[key])

        # Get mode
        if inpt is None:
            W = attn_proj.weight
            W = einops.rearrange(W, f"{their_shape} -> {my_shape}", **sizes)
            return W

        # Set mode
        W = einops.rearrange(inpt, f"{my_shape} -> {their_shape}", **sizes)
        update_param(attn_proj, "weight", W)

    def roberta_qkv_bias(layer, key: str, inpt: Optional[Any]=None):
        # Prepare shape changing
        their_shape = "(n_heads d_head)"
        my_shape    = "n_heads d_head"
        sizes = generate_sizes_dict(my_shape, cfg)

        # Get attn proj module
        attn = layer.attention
        attn_proj = get_attrs(attn, attn_proj_map[key])

        if inpt is None:
            b = attn_proj.bias
            b = einops.rearrange(b, f"{their_shape} -> {my_shape}", **sizes)
            return b

        # Set mode
        b = einops.rearrange(inpt, f"{my_shape} -> {their_shape}", **sizes)
        update_param(attn_proj, "bias", b)


    roberta_layer_map = {
        "attn.ln_out"           : "attention.output.LayerNorm",
        "attn.ln_out.w"         : "attention.output.LayerNorm.weight",
        "attn.ln_out.b"         : "attention.output.LayerNorm.bias",

        "attn"          : "attention",
        "attn.q_proj"   : "attention.self.query",
        "attn.k_proj"   : "attention.self.key",
        "attn.v_proj"   : "attention.self.value",

        **generate_attn_qkv_functions(roberta_qkv_weight, roberta_qkv_bias),

        "attn.out"      : "attention.output",
        "attn.out_proj" : "attention.output.dense",
        "attn.W_O"      : "attention.output.dense.weight",
        "attn.b_O"      : "attention.output.dense.bias",

        "attn.inv_out_proj" : "attention.inv_out_proj",
        "attn.W_O_inv"  : "attention.inv_out_proj.weight",
        "attn.b_O_inv"  : "attention.inv_out_proj.inverse_bias",

        "mlp.ln_out"           : "output.LayerNorm",
        "mlp.ln_out.w"         : "output.LayerNorm.weight",
        "mlp.ln_out.b"         : "output.LayerNorm.bias",

        # "mlp"           : "intermediate",
        "mlp.in_proj"   : "intermediate.dense",
        "mlp.W_in"      : "intermediate.dense.weight",
        "mlp.b_in"      : "intermediate.dense.bias",

        "activation_fn" : "intermediate.intermediate_act_fn",

        "mlp.out_proj"  : "output.dense",
        "mlp.W_out"     : "output.dense.weight",
        "mlp.b_out"     : "output.dense.bias",
    }
    return roberta_layer_map

#####################################################################################
# VISION TRANSFORMERS (eg: ViT)
#####################################################################################

# ViT
#####

vit_model_map = {
    "model"           : "vit",
    "layers"          : "vit.encoder.layer",
    "embed"           : "vit.embeddings",
    #"embed.W_embed"   : "vit.embeddings.patch_embeddings.weight", ? is CNN
    #"pos_embed.W_pos" : Can't see any ???
    "ln_final"        : "layernorm",
    "ln_final.w"      : "layernorm.weight",
    "ln_final.b"      : "layernorm.bias",
    "unembed"         : "classifier",
    "unembed.W_U"     : "classifier.weight",
    "unembed.b_U"     : "classifier.bias",
}

def build_vit_layer_map(cfg: ConfigClass):
    attn_proj_map = {
        "q": "attention.query",
        "k": "attention.key",
        "v": "attention.value",
        "o": "output.dense"
    }

    def vit_qkv_weight(layer, key: str, inpt: Optional[Any]=None):
        # Prepare shape changing
        their_shape = "(n_heads d_head) d_model"
        my_shape    = "n_heads d_head d_model"
        sizes = generate_sizes_dict(my_shape, cfg)

        # Get attn proj module
        attn = layer.attention
        attn_proj = get_attrs(attn, attn_proj_map[key])

        # Get mode
        if inpt is None:
            W = attn_proj.weight
            W = einops.rearrange(W, f"{their_shape} -> {my_shape}", **sizes)
            return W

        # Set mode
        W = einops.rearrange(inpt, f"{my_shape} -> {their_shape}", **sizes)
        update_param(attn_proj, "weight", W)

    def vit_qkv_bias(layer, key: str, inpt: Optional[Any]=None):
        # Prepare shape changing
        their_shape = "(n_heads d_head)"
        my_shape    = "n_heads d_head"
        sizes = generate_sizes_dict(my_shape, cfg)

        # Get attn proj module
        attn = layer.attention
        attn_proj = get_attrs(attn, attn_proj_map[key])

        if inpt is None:
            b = attn_proj.bias
            b = einops.rearrange(b, f"{their_shape} -> {my_shape}", **sizes)
            return b

        # Set mode
        b = einops.rearrange(inpt, f"{my_shape} -> {their_shape}", **sizes)
        update_param(attn_proj, "bias", b)


    vit_layer_map = {
        "attn.ln_in"    : "layernorm_before",
        "attn.ln_in.w"  : "layernorm_before.weight",
        "attn.ln_in.b"  : "layernorm_before.bias",

        "attn"          : "attention",
        "attn.q_proj"   : "attention.attention.query",
        "attn.k_proj"   : "attention.attention.key",
        "attn.v_proj"   : "attention.attention.value",

        **generate_attn_qkv_functions(vit_qkv_weight, vit_qkv_bias),

        "attn.out"      : "attention.output",
        "attn.out_proj" : "attention.output.dense",
        "attn.W_O"      : "attention.output.dense.weight",
        "attn.b_O"      : "attention.output.dense.bias",

        "attn.inv_out_proj" : "attention.inv_out_proj",
        "attn.W_O_inv"  : "attention.inv_out_proj.weight",
        "attn.b_O_inv"  : "attention.inv_out_proj.inverse_bias",

        "mlp.ln_in"     : "layernorm_after",
        "mlp.ln_in.w"   : "layernorm_after.weight",
        "mlp.ln_in.b"   : "layernorm_after.bias",

        # "mlp"           : "intermediate",
        "mlp.in_proj"   : "intermediate.dense",
        "mlp.W_in"      : "intermediate.dense.weight",
        "mlp.b_in"      : "intermediate.dense.bias",

        "activation_fn" : "intermediate.intermediate_act_fn",

        "mlp.out_proj"  : "output.dense",
        "mlp.W_out"     : "output.dense.weight",
        "mlp.b_out"     : "output.dense.bias",
    }

    return vit_layer_map

#####################################################################################
# Encoder + Decoder Models
#####################################################################################

# T5 Model Map for PyTorch T5 Model
###################################

t5_model_map = {
    "model"           : "t5",
    "layers"          : "encoder.block",
    "embed"           : "shared",
    "embed.W_E"       : "shared.weight",
    "pos_embed.W_pos" : "encoder.embed_tokens.position_embeddings.weight",
    "ln_final"        : "encoder.final_layer_norm",
    "ln_final.w"      : "encoder.final_layer_norm.weight",
    "ln_final.b"      : "encoder.final_layer_norm.bias",
    "unembed"         : "lm_head",
    "unembed.W_U"     : "lm_head.weight",
    "unembed.b_U"     : None,
}

def build_t5_layer_map(cfg: ConfigClass):
    attn_proj_map = {
        "q": "self_attn.q",
        "k": "self_attn.k",
        "v": "self_attn.v",
        "o": "self_attn.o"
    }

    def t5_qkv_weight(layer, key: str, inpt: Optional[Any]=None):
        # Prepare shape changing
        their_shape = "(n_heads, d_head) d_model"
        my_shape    = "n_heads, d_head, d_model"
        sizes = generate_sizes_dict(my_shape, cfg)

        # Get attn proj module
        attn = layer.self_attn
        attn_proj = get_attrs(attn, attn_proj_map[key])

        # Get mode
        if inpt is None:
            W = attn_proj.weight
            W = einops.rearrange(W, f"{their_shape} -> {my_shape}", **sizes)
            return W

        # Set mode
        W = einops.rearrange(inpt, f"{my_shape} -> {their_shape}", **sizes)
        update_param(attn_proj, "weight", W)

    def t5_qkv_bias(layer, key: str, inpt: Optional[Any]=None):
        # Prepare shape changing
        their_shape = "(n_heads, d_head)"
        my_shape    = "n_heads, d_head"
        sizes = generate_sizes_dict(my_shape, cfg)

        # Get attn proj module
        attn = layer.self_attn
        attn_proj = get_attrs(attn, attn_proj_map[key])

        if inpt is None:
            b = attn_proj.bias
            b = einops.rearrange(b, f"{their_shape} -> {my_shape}", **sizes)
            return b

        # Set mode
        b = einops.rearrange(inpt, f"{my_shape} -> {their_shape}", **sizes)
        update_param(attn_proj, "bias", b)


    t5_layer_map = {
        "attn.ln_in"    : "self_attn.layer_norm",
        "attn.ln_in.w"  : "self_attn.layer_norm.weight",
        "attn.ln_in.b"  : "self_attn.layer_norm.bias",

        "attn"          : "self_attn",
        "attn.q_proj"   : "self_attn.q",
        "attn.k_proj"   : "self_attn.k",
        "attn.v_proj"   : "self_attn.v",

        **generate_attn_qkv_functions(t5_qkv_weight, t5_qkv_bias),

        "attn.out_proj" : "self_attn.o",
        "attn.W_O"      : "self_attn.o.weight",
        "attn.b_O"      : "self_attn.o.bias",

        "mlp.ln_in"     : "fc.layer_norm",
        "mlp.ln_in.w"   : "fc.layer_norm.weight",
        "mlp.ln_in.b"   : "fc.layer_norm.bias",

        "mlp"           : "fc",
        "mlp.in_proj"   : "fc.DenseReluDense.wi",
        "mlp.W_in"      : "fc.DenseReluDense.wi.weight",
        "mlp.b_in"      : "fc.DenseReluDense.wi.bias",

        "activation_fn" : "fc.act",

        "mlp.out_proj"  : "fc.DenseReluDense.wo",
        "mlp.W_out"     : "fc.DenseReluDense.wo.weight",
        "mlp.b_out"     : "fc.DenseReluDense.wo.bias",
    }
    return t5_layer_map


#####################################################################################
# Build Model Layer Map interfaces
#####################################################################################

# Define Helper Functions
#########################

def get_attrs(obj, attr_string):
    nested_attributes = attr_string.split('.')
    current_attr = obj
    for attr_name in nested_attributes:
        current_attr = getattr(current_attr, attr_name)
    return current_attr

def set_attrs(obj, attr_string, value, override=True):
    nested_attrs = attr_string.split('.')
    nested_attrs, final_attr = nested_attrs[:-1], nested_attrs[-1]
    current_attr = get_attrs(obj, ".".join(nested_attrs)) if len(nested_attrs) > 0 else obj
    if not override and hasattr(current_attr, final_attr):
        return
    setattr(current_attr, final_attr, value)

def get_model_key_map(config: ConfigClass):
    architecture = config.architecture
    if architecture == "OPTForCausalLM":
        return opt_model_map
    if architecture in ["LLaMAForCausalLM", "LlamaForCausalLM"]:
        return llama_model_map
    if architecture == "MistralForCausalLM":
        return mistral_model_map
    if architecture == "GemmaForCausalLM":
        return gemma_model_map
    if architecture == "Gemma2ForCausalLM":
        return gemma2_model_map
    if architecture == "Gemma3ForConditionalGeneration":
        return gemma3_model_map
    if architecture == "Gemma3ForCausalLM":
        return gemma3_textonly_model_map
    if architecture == "PhiForCausalLM":
        return phi_model_map
    if architecture == "Phi3ForCausalLM":
        return phi3_model_map
    if architecture == "GPTNeoXForCausalLM":
        return gpt_neox_model_map
    if architecture == "GPT2LMHeadModel":
        return gpt2_model_map
    if architecture == "RobertaForMaskedLM":
        return roberta_model_map
    if architecture == "ViTForImageClassification":
        return vit_model_map

    raise NotImplementedError(f"Architecture {architecture} not implemented")

def get_layer_key_map(config: ConfigClass):
    architecture = config.architecture

    if architecture == "OPTForCausalLM":
        return build_opt_layer_map(config)
    if architecture in ["LLaMAForCausalLM", "LlamaForCausalLM"]:
        return build_llama_layer_map(config)
    if architecture == "MistralForCausalLM":
        return build_mistral_layer_map(config)
    if architecture == "GemmaForCausalLM":
        return build_gemma_layer_map(config)
    if architecture == "Gemma2ForCausalLM":
        return build_gemma2_layer_map(config)
    if architecture == "Gemma3ForCausalLM":
        return build_gemma3_layer_map(config)
    if architecture == "Gemma3ForConditionalGeneration":
        return build_gemma3_layer_map(config)
    if architecture == "PhiForCausalLM":
        return build_phi_layer_map(config)
    if architecture == "Phi3ForCausalLM":
        return build_phi3_layer_map(config)
    if architecture == "GPTNeoXForCausalLM":
        return build_gpt_neox_layer_map(config)
    if architecture == "GPT2LMHeadModel":
        return build_gpt2_layer_map(config)
    if architecture == "RobertaForMaskedLM":
        return build_roberta_layer_map(config)
    if architecture == "ViTForImageClassification":
        return build_vit_layer_map(config)

    raise NotImplementedError(f"Architecture {architecture} not implemented")


# Define Real Model Map and Layer Maps
######################################

class ModelMap:
    def __init__(self, model, cfg):
        self.cfg         = cfg
        self.model       = model
        self.key_map     = get_model_key_map(cfg)

        # Handle layers
        self.orig_layers = self["layers"]
        self.layers = [
            ModelLayerMap(self.cfg, layer) for layer in self.orig_layers
        ]

    def __getitem__(self, __name: str):
        key = self.key_map[__name]
        return get_attrs(self.model, key)

    def __setitem__(self, key, inpt):
        keys = key.split('.')
        attr = keys[-1]
        module = get_attrs(self.model, ".".join(keys[:-1]))
        params = module.state_dict()
        params[attr] = inpt
        module.load_state_dict(params)

class ModelLayerMap:
    def __init__(self, cfg, layer):
        self.cfg   = cfg
        self.layer = layer
        self.key_map = get_layer_key_map(cfg)

    @property
    def names(self):
        return list(self.key_map.keys())

    def __contains__(self, __name):
        return (__name in self.key_map)

    def __getitem__(self, __name):
        key = self.key_map[__name]

        if isinstance(key, str):
            if key == "layer":
                return self.layer
            if "inv_out_proj" in key:
                return NotImplementedError()
            return get_attrs(self.layer, key)

        if isinstance(key, Callable):
            return key(self.layer)

    def __setitem__(self, __name: str, __value: Any) -> None:
        key = self.key_map[__name]
        if isinstance(key, Callable):
            return key(self.layer, __value)

        if key is None:
            return None

        if not isinstance(key, str):
            raise ValueError("Invalid key, must be string or callable")

        # Get the module and attribute name
        keys = key.split('.')
        module = get_attrs(self.layer, ".".join(keys[:-1]))
        attr   = keys[-1]

        if attr == "inv_out_proj":
            setattr(module, attr, __value)
            return

        # If setting an attribute of a module (eg: weights or biases), update
        params = module.state_dict()
        params[attr] = __value
        module.load_state_dict(params)
        return

    def __str__(self):
        out_str  = "Wrapper for Transformer Layer\n"
        out_str += self.key_map.keys().__str__()
        out_str += "\nOriginal Layer Structure:\n"
        out_str += self.layer.__str__()
        return out_str

    def __getattr__(self, __name):
        key = self.key_map[__name]

        # Find all names that start with __name followed by a dot
        prefix = __name + "."
        remaining_names = sorted([n.split(prefix, 1)[1] for n in self.names if n.startswith(prefix)])

        if isinstance(key, str):
            mod = get_attrs(self.layer, key)
            for n in remaining_names:
                set_attrs(mod, n, self[f"{__name}.{n}"], override=False)
            return mod

        if isinstance(key, Callable):
            return key(self.layer)



from typing import Optional, List

import numpy as np
import torch
import wandb
import copy
import gc

from .model import Model
from .data_classes import PruningConfig, RunDataHistory, \
                          RunDataItem, ActivationOverview
from .eval import evaluate_all
from .scoring import score_indices_by, score_indices
from .activations import get_midlayer_data, get_top_frac, \
    choose_attn_heads_by, save_timestamped_tensor_dict
from .texts import prepare

def prune_and_evaluate(
        opt: Model,
        pruning_config: PruningConfig,
        focus_out: Optional[dict] = None,
        cripple_out: Optional[dict] = None,
        iteration: Optional[int] = None,
    ):
    """
    Prune and evaluate the model

    Args:
        opt (Model): model to prune and evaluate
        pruning_config (PruningConfig): config for pruning
        focus_out (dict): output of get_midlayer_data for focus dataset
        cripple_out (dict): output of get_midlayer_data for cripple dataset
        iteration (int): iteration number for when activations are not recalculated

    Returns:
        output (RunDataItem): Eval data to add to RunDataHistory.
    """
    c = copy.deepcopy(pruning_config)

    # Find out what we are doing
    do_ff   = pruning_config.ff_frac > 0
    do_attn = pruning_config.attn_frac > 0
    if not do_ff and not do_attn:
        raise ValueError("Must prune at least one of FF or Attention")
    if do_attn and pruning_config.attn_mode not in ["pre-out", "value"]:
        raise NotImplementedError("attn_mode must be 'pre-out' or 'value'")

    # Get midlayer activations of FF and ATTN
    if pruning_config.recalculate_activations:
        focus_out   = get_midlayer_data( opt, pruning_config.focus,
            pruning_config.collection_sample_size, pruning_config.attn_mode )
        cripple_out = get_midlayer_data( opt, pruning_config.cripple,
            pruning_config.collection_sample_size, pruning_config.attn_mode )

    # Otherwise, import activation data, and adjust the "pruning fraction"
    else:
        c["ff_frac"]   = min( 1.0, c["ff_frac"]*(iteration+1) )
        c["attn_frac"] = min( 1.0, c["attn_frac"]*(iteration+1) )
        assert not (focus_out is None or cripple_out is None or iteration is None), \
            "Must provide focus_out and cripple_out if not recalculate_activations"
    # Prune the model using the activation data
    data = score_and_prune(opt, focus_out, cripple_out, c, c.save)

    # Evaluate the model
    with torch.no_grad():
        opt.hooks.disable_all_collect_hooks()
        gc.collect()
        torch.cuda.empty_cache()
        eval_out = evaluate_all(opt, c.eval_sample_size, c.datasets,
                                dataset_tokens_to_skip=c.collection_sample_size,
                                config_args={'batch_size': 4})
        data.update(eval_out)

    return data

def score_and_prune( opt: Model,
            focus_activations_data: ActivationOverview,
            cripple_activations_data: ActivationOverview,
            pruning_config: PruningConfig,
            save=False,
        ):
    # Get the top fraction FF activations and prune
    ff_frac, ff_eps     = pruning_config.ff_frac,   pruning_config.ff_eps
    attn_frac, attn_eps = pruning_config.attn_frac, pruning_config.attn_eps
    do_ff   = ff_frac > 0
    do_attn = attn_frac > 0

    act_subset = pruning_config.scoring_normalization
    if do_ff > 0:
        ff_focus_data   = focus_activations_data.mlp[act_subset]
        ff_cripple_data = cripple_activations_data.mlp[act_subset]
        ff_scoring_fn = score_indices_by(pruning_config.ff_scoring)

        ff_scores = ff_scoring_fn(opt, ff_focus_data, ff_cripple_data, ff_eps)
        ff_criteria, ff_threshold = get_top_frac(ff_scores, ff_frac)
        opt.hooks.delete_mlp_neurons(ff_criteria)

    # Get the top fraction of Attention activations and prune
    if do_attn > 0:
        attn_focus_data   = focus_activations_data.attn[act_subset]
        attn_cripple_data = cripple_activations_data.attn[act_subset]
        # scoring for attention
        attn_scoring_fn = score_indices_by(pruning_config.attn_scoring)
        attn_scores = attn_scoring_fn(opt, attn_focus_data, attn_cripple_data, attn_eps)

        # offset by means if desired (probably bad?)
        means = None
        if pruning_config.do_attn_mean_offset:
            means = attn_focus_data["mean"]

        # get criteria for "neurons", or for "heads" if using full heads
        if pruning_config.attn_prune_heads:
            attn_head_scoring_fn = \
                choose_attn_heads_by(pruning_config.attn_prune_heads_mode)
            attn_criteria, attn_threshold = \
                attn_head_scoring_fn(opt, attn_scores, attn_frac)
            attn_criteria = opt.expand_remove_heads_to_remove_indices(attn_criteria)
        else:
            attn_criteria, attn_threshold = get_top_frac(attn_scores, attn_frac)
            _shape = (opt.cfg.n_layers, opt.cfg.n_heads, opt.cfg.d_head)
            attn_criteria = attn_criteria.reshape(_shape)

        # get criteria and prune if using only attention neurons
        if pruning_config.attn_mode == "pre-out":
            opt.hooks.delete_attn_neurons( attn_criteria) # TODO: add option for means
        elif pruning_config.attn_mode == "value":
            raise NotImplementedError("'value' pruning not yet implemented in v1")
            opt.delete_attn_values( attn_criteria, means )
        else:
            raise NotImplementedError("attn_mode must be 'pre-out' or 'value'")

    # Save the removals to file
    tensor_data = {
        "ff_scores": ff_scores if do_ff else None,
        # FIXME: doesn't return attn_std_mean
        "attn_scores": attn_scores if do_attn else None,
        "ff_criteria": ff_criteria if do_ff else None,
        "attn_criteria": attn_criteria if do_attn else None,
    }

    if save:
        subdirectory = f"{pruning_config.save_subdirectory}/" or ""
        path = f"{subdirectory}saved_tensors/{opt.model_size}"
        filename = f"{pruning_config.cripple}-{pruning_config.focus}-{opt.model_size}-{pruning_config.ff_frac}-recent.pt"
        save_timestamped_tensor_dict( opt, tensor_data, "activation_metrics", path, filename )

    # Initialize the output dictionary
    data = RunDataItem()

    data.update({'deletions': {
        "ff_threshold": ff_threshold if do_ff else 0,
        "attn_threshold": attn_threshold if do_attn else 0,
        "ff_del": float( torch.sum(ff_criteria) ) if do_ff else 0,
        "attn_del": float( torch.sum(attn_criteria) ) if do_attn else 0,
    }})

    data.update({'deletions_per_layer': {
        'ff': ff_criteria.sum(dim=-1).tolist() if do_ff else [],
        'attn': attn_criteria.sum(dim=-1).tolist() if do_attn else [],
    }})

    # Save removals and scores to history
    def _numpify(x):
        if x is None:
            return None
        if x.dtype == torch.bfloat16:
            return x.float().cpu().numpy()
        return x.cpu().numpy()

    data.update({'raw': {
        k: _numpify(v) for k,v in tensor_data.items()
    }})

    return data

def prune_random( opt: Model,
        ff_frac: float,
        attn_frac: float,
        ff_pruned: Optional[np.ndarray] = None,
        attn_pruned: Optional[np.ndarray] = None,
        ):
    """Prune a random fraction of FF and Attention weights
    Args:
        opt (Model): model to prune and evaluate
        ff_frac (float): fraction of FF to prune
        attn_frac (float): fraction of Attention to prune
        ff_pruned: list of which mlp neurons have already been pruned
        attn_pruned: list of which attn neurons have already been pruned

    Returns:
        ff_pruned: updated list of which mlp neurons have already been pruned
        attn_pruned: updated list of which attn neurons have already been pruned
        data_out: summary info on which neurons have been pruned
    """
    if ff_pruned is None:
        ff_pruned = np.zeros( (opt.cfg.n_layers, opt.cfg.d_mlp), dtype=np.bool_ )
    if attn_pruned is None:
        attn_pruned = np.zeros( (opt.cfg.n_layers, opt.cfg.d_model ), dtype=np.bool_ )

    n_ff_to_prune   = int( ff_frac   * opt.cfg.d_mlp )
    n_attn_to_prune = int( attn_frac * opt.cfg.d_model )

    # First prune the FF
    if not ff_frac == 0:
        for layer in range( opt.cfg.n_layers ):
            # choose new ff neurons to prune
            indices = np.where(ff_pruned[layer] == 0)[0]
            random_indices = np.random.choice(indices, n_ff_to_prune, replace=False)
            ff_pruned[layer][random_indices] = 1

        # Prune the model
        opt.hooks.delete_mlp_neurons( torch.tensor(ff_pruned) )

    if not attn_frac == 0:
        for layer in range( opt.cfg.n_layers ):
            # choose new attention heads to prune
            indices = np.where(attn_pruned[layer] == 0)[0]
            random_indices = np.random.choice(indices, n_attn_to_prune, replace=False)
            attn_pruned[layer][random_indices] = 1

        # Prune the model
        opt.hooks.delete_attn_neurons( torch.tensor(attn_pruned) )

    data_out = {
        "ff_del": n_ff_to_prune*opt.cfg.n_layers,
        "attn_del": n_attn_to_prune*opt.cfg.n_layers
    }
    return ff_pruned, attn_pruned, data_out

def prune_random_and_evaluate( opt: Model,
        c: PruningConfig,
        ff_pruned: Optional[np.ndarray] = None,
        attn_pruned: Optional[np.ndarray] = None,
        ):
    """
    To use, run once with ff_pruned=None and attn_pruned=None, then run again
    with the parameters given as output passed back in.

    Args:
        opt (Model): The model to prune and evaluate
        c (PruningConfig): The pruning configuration
        ff_pruned (Optional[np.ndarray]): Bool list of FF neurons, default None.
        attn_pruned (Optional[np.ndarray], optional: Bool list of ATTN neurons, default None.

    Returns:
        ff_pruned (Optional[np.ndarray]):
        attn_pruned (Optional[np.ndarray]):
        data (RunDataItem):
    """


    # Prune the model randomly
    ff_pruned, attn_pruned, data_out = \
        prune_random( opt, c.ff_frac, c.attn_frac, ff_pruned, attn_pruned )

    # Initialize the output dictionary
    data = RunDataItem()

    # Evaluate the model
    data.update(
        evaluate_all( opt, c.eval_sample_size, c.datasets,
                      dataset_tokens_to_skip=c.collection_sample_size )
    )

    data.update({'deletions': data_out })

    data.update({'deletions_per_layer': {
        'ff': ff_pruned.sum(axis=-1).tolist() if (not ff_pruned is None) else 0,
        'attn': attn_pruned.sum(axis=-1).tolist() if (not attn_pruned is None) else 0,
    }})

    return ff_pruned, attn_pruned, data

######################################################################################
# Run Whole Pruning Procedure from Config
######################################################################################

def run_pruning(c: PruningConfig):
    # Initialise Model and show details about model
    opt = Model(
        c.model_size,
        limit=c.token_limit,
        dtype=c.dtype,
        svd_attn=c.svd_attn,
        use_accelerator=c.use_accelerator,
        model_device=c.model_device,
        mask_fn=c.mask_fn,
        )

    # Prepare data logging
    history = RunDataHistory(c.datasets)
    wandb.init(
        project=c.wandb_project,
        entity=c.wandb_entity,
        name=c.wandb_run_name,
        )
    wandb.config.update(c.to_dict(), allow_val_change=True)

    # Evaluate model before removal of any neurons
    if c.run_pre_test:
        data = evaluate_all(opt, c.eval_sample_size,
            c.datasets, c.collection_sample_size)
        history.add(data)
        print(history.df.T)

    # If pruning randomly, no need to get activations
    if c.ff_scoring == "random" and c.attn_scoring == "random":
        ff_pruned, attn_pruned = None, None
        for i in range(c.n_steps):
            ff_pruned, attn_pruned, data = \
                prune_random_and_evaluate(opt, c, ff_pruned, attn_pruned)
            history.add(data)

    # Iteratively prune neurons and evaluate
    elif c.recalculate_activations:
        for _ in range(c.n_steps):
            data = prune_and_evaluate(opt, c)
            history.add(data)

    # Non-iteratively get activations, then iteratively prune and evaluate
    else:
        focus_out   = get_midlayer_data(opt, c.focus,
                        c.collection_sample_size, c.attn_mode)
        cripple_out = get_midlayer_data(opt, c.cripple,
                        c.collection_sample_size, c.attn_mode)
        for i in range(c.n_steps):
            data = prune_and_evaluate(opt, c, focus_out, cripple_out, i)
            history.add(data)

    # Format history to print
    print(history.history[-1])
    print(history.df.T)
    print(history.df.T.to_csv())

    return opt, history

######################################################################################
# "Forsaken"-style pruning
######################################################################################

def forsaken_pruning(c: PruningConfig,
        num_texts: int = 1,
        lr: float = 0.1,
        sigmoid_offset: float = 2.0,
        l1_norm_coeff: float = 1.0,
        ):
    # Initilaise Model and show details about model
    c.mask_fn = "sigmoid"
    c.misc = {
        "num_texts": num_texts,
        "lr": lr,
        "sigmoid_offset": sigmoid_offset,
        "l1_norm_coeff": l1_norm_coeff,
    }

    opt = Model(
        c.model_size,
        limit=c.token_limit,
        dtype=c.dtype,
        svd_attn=c.svd_attn,
        use_accelerator=c.use_accelerator,
        model_device=c.model_device,
        mask_fn=c.mask_fn,
        )

    # Prepare data logging
    history = RunDataHistory(c.datasets)
    wandb.init(
        project=c.wandb_project,
        entity=c.wandb_entity,
        name=c.wandb_run_name,
        )
    wandb.config.update(c.to_dict())

    # Evaluate model before removal of any neurons
    if c.run_pre_test:
        data = evaluate_all(opt, c.eval_sample_size,
            c.datasets, c.collection_sample_size)
        history.add(data)
        print(history.df.T)

    # Get activations
    focus_out   = get_midlayer_data(opt, c.focus,
                    c.collection_sample_size, c.attn_mode)
    cripple_out = get_midlayer_data(opt, c.cripple,
                    c.collection_sample_size, c.attn_mode)

    def normalize_scores(scores):
        normed_scores = []
        for score in scores:
            normed_scores.append(
                (score - score.mean()) / score.std()
            )
        return torch.stack(normed_scores)

    # Set masks for feed-forward layers
    ff_scores   = score_indices(c.ff_scoring,
        opt, focus_out.mlp.orig,   cripple_out.mlp.orig)
    ff_masks    = sigmoid_offset - normalize_scores(ff_scores)
    for layer_index in range(opt.cfg.n_layers):
        mask = opt.masks["mlp_pre_out"][layer_index]
        mask.set_mask(ff_masks[layer_index])

    # Set masks for attention heads
    attn_scores = score_indices(c.attn_scoring,
        opt, focus_out.attn.orig, cripple_out.attn.orig)
    attn_masks  = sigmoid_offset - normalize_scores(attn_scores)
    for layer_index in range(opt.cfg.n_layers):
        mask = opt.masks["attn_pre_out"][layer_index]
        mask.set_mask(attn_masks[layer_index])

    # Evaluate again now that we have adjusted the masks
    if True:
        data = evaluate_all(opt, c.eval_sample_size,
            c.datasets, c.collection_sample_size)
        history.add(data)

    # Get parameters for back propagation
    mask_params = [
        *[p for mask in opt.masks["mlp_pre_out"] for p in mask.parameters()],
        *[p for mask in opt.masks["attn_pre_out"] for p in mask.parameters()],
    ]
    mask_l1_norm = torch.stack([
        *[(1-mask.get_mask()).mean() for mask in opt.masks["mlp_pre_out"]],
        *[(1-mask.get_mask()).mean() for mask in opt.masks["attn_pre_out"]],
    ]).mean()

    # Generate Inputs
    n_iter = 4
    optim = torch.optim.LBFGS(mask_params, lr, max_iter=n_iter)
    kl_loss_fn = torch.nn.KLDivLoss()
    #ce_loss_fn = torch.nn.CrossEntropyLoss()

    # Load datasets
    def gen_texts(num_texts=1):
        _cripple_texts, _focus_texts = [], []

        cripple_dataset, cripple_label, _skip50 = prepare(c.cripple)
        i = 0
        for data in cripple_dataset:
            i += 1
            if i > num_texts:
                break
            _cripple_texts.append(data[cripple_label])

        focus_dataset, focus_label, _skip50     = prepare(c.focus)
        i = 0
        for data in focus_dataset:
            i += 1
            if i > num_texts:
                break
            _focus_texts.append(data[focus_label])

        return _cripple_texts, _focus_texts

    # Begin calculating loss for with LBGFS
    def get_new_ids(n_batches = None):
        batches = []
        cripple_texts, focus_texts = gen_texts()
        bad_ids, junk_ids, good_ids = [], [], []
        with torch.no_grad():
            for text in cripple_texts:
                bad_ids.append( opt.get_ids(text) )
                junk_ids.append(
                    torch.randint_like(bad_ids[-1], 5, opt.tokenizer.vocab_size)
                )
            for text in focus_texts:
                good_ids.append( opt.get_ids(text) )
        return bad_ids, junk_ids, good_ids

    for j in range(c.n_steps//n_iter):
        bad_ids, junk_ids, good_ids = get_new_ids()

        # Begin LBGFS
        def closure():
            loss = 0
            optim.zero_grad()

            # Generate loss
            loss += mask_l1_norm * l1_norm_coeff

            for i in range(num_texts):
                # Get junk loss L_kl(gamma,P)
                with torch.no_grad():
                    junk_logits = opt.get_logits(input_ids=junk_ids[i])[..., :-1, :]
                bad_logits = opt.get_logits(input_ids=bad_ids[i])[..., :-1, :]
                loss += kl_loss_fn(bad_logits, junk_logits).mean()

                # Get good loss L_kl(gamma,Q)
                with torch.no_grad():
                    opt.masking_enabled = False
                    orig_logits = opt.get_logits(input_ids=good_ids[i])[..., :-1, :]
                    opt.masking_enabled = True
                new_logits = opt.get_logits(input_ids=good_ids[i])[..., :-1, :]
                loss += kl_loss_fn(new_logits, orig_logits).mean()

            # Backpropagate
            loss.backward(retain_graph=True)
            return loss

        # loss step
        for i in range(n_iter):
            optim.step(closure)
            data = evaluate_all(opt, c.eval_sample_size,
                c.datasets, c.collection_sample_size)
            history.add(data)

    # Format history to print
    print(history.history[-1])
    print(history.df.T)
    print(history.df.T.to_csv())

    return opt, history




#!/usr/bin/env python3
"""Test the exact example provided by the user."""

import sys
import os
sys.path.append(os.path.join(os.path.dirname(__file__), 'src'))

import torch
from neo_taker import Model

def test_user_example(model_repo: str = "nickypro/tinyllama-15m"):
    """Test the exact example from the user requirements."""
    print("Testing the exact user example...")
    
    # Create model
    model = Model(
        model_repo=model_repo,
        model_device="cpu"
    )
    
    # Define hook function exactly as specified
    def hook(act, hook):
        print(hook.name)
        act[-1, -1, :] = 0
        return act

    # Test the exact syntax from user requirements
    tokens = model.to_tokens("Hello world!")
    
    print(f"Input tokens shape: {tokens.shape}")
    
    # Use the exact hook syntax specified by the user
    with model.hooks(fwd_hooks=[("blocks.0.hook_resid_pre", hook)]):
        result = model(tokens, return_type="logits")
    
    print(f"Output shape: {result.shape}")
    print("✓ User example works perfectly!")
    
    # Test a few more hook points from the user's list
    hook_points_to_test = [
        "hook_embed",
        "blocks.0.ln1.hook_scale", 
        "blocks.0.attn.hook_q",
        "blocks.0.mlp.hook_pre",
        "ln_final.hook_normalized"
    ]
    
    print(f"\nTesting additional hook points...")
    for hook_point in hook_points_to_test:
        try:
            with model.hooks(fwd_hooks=[(hook_point, hook)]):
                result = model(tokens, return_type="logits")
            print(f"✓ {hook_point}")
        except Exception as e:
            print(f"✗ {hook_point}: {e}")
    
    print(f"\n🎉 All tests passed! neo_taker now supports TransformerLens-style hooks!")

if __name__ == "__main__":
    model_repo = "nickypro/tinyllama-15m"
    if len(sys.argv) > 1:
        model_repo = sys.argv[1]
    test_user_example(model_repo)




import einops

from transformer_lens.HookedTransformerConfig import HookedTransformerConfig


def convert_bert_weights(bert, cfg: HookedTransformerConfig):
    embeddings = bert.bert.embeddings
    state_dict = {
        "embed.embed.W_E": embeddings.word_embeddings.weight,
        "embed.pos_embed.W_pos": embeddings.position_embeddings.weight,
        "embed.token_type_embed.W_token_type": embeddings.token_type_embeddings.weight,
        "embed.ln.w": embeddings.LayerNorm.weight,
        "embed.ln.b": embeddings.LayerNorm.bias,
    }

    for l in range(cfg.n_layers):
        block = bert.bert.encoder.layer[l]
        state_dict[f"blocks.{l}.attn.W_Q"] = einops.rearrange(
            block.attention.self.query.weight, "(i h) m -> i m h", i=cfg.n_heads
        )
        state_dict[f"blocks.{l}.attn.b_Q"] = einops.rearrange(
            block.attention.self.query.bias, "(i h) -> i h", i=cfg.n_heads
        )
        state_dict[f"blocks.{l}.attn.W_K"] = einops.rearrange(
            block.attention.self.key.weight, "(i h) m -> i m h", i=cfg.n_heads
        )
        state_dict[f"blocks.{l}.attn.b_K"] = einops.rearrange(
            block.attention.self.key.bias, "(i h) -> i h", i=cfg.n_heads
        )
        state_dict[f"blocks.{l}.attn.W_V"] = einops.rearrange(
            block.attention.self.value.weight, "(i h) m -> i m h", i=cfg.n_heads
        )
        state_dict[f"blocks.{l}.attn.b_V"] = einops.rearrange(
            block.attention.self.value.bias, "(i h) -> i h", i=cfg.n_heads
        )
        state_dict[f"blocks.{l}.attn.W_O"] = einops.rearrange(
            block.attention.output.dense.weight,
            "m (i h) -> i h m",
            i=cfg.n_heads,
        )
        state_dict[f"blocks.{l}.attn.b_O"] = block.attention.output.dense.bias
        state_dict[f"blocks.{l}.ln1.w"] = block.attention.output.LayerNorm.weight
        state_dict[f"blocks.{l}.ln1.b"] = block.attention.output.LayerNorm.bias
        state_dict[f"blocks.{l}.mlp.W_in"] = einops.rearrange(
            block.intermediate.dense.weight, "mlp model -> model mlp"
        )
        state_dict[f"blocks.{l}.mlp.b_in"] = block.intermediate.dense.bias
        state_dict[f"blocks.{l}.mlp.W_out"] = einops.rearrange(
            block.output.dense.weight, "model mlp -> mlp model"
        )
        state_dict[f"blocks.{l}.mlp.b_out"] = block.output.dense.bias
        state_dict[f"blocks.{l}.ln2.w"] = block.output.LayerNorm.weight
        state_dict[f"blocks.{l}.ln2.b"] = block.output.LayerNorm.bias

    pooler = bert.bert.pooler
    state_dict["pooler.W"] = pooler.dense.weight.T
    state_dict["pooler.b"] = pooler.dense.bias

    mlm_head = bert.cls.predictions
    state_dict["mlm_head.W"] = mlm_head.transform.dense.weight.T
    state_dict["mlm_head.b"] = mlm_head.transform.dense.bias
    state_dict["mlm_head.ln.w"] = mlm_head.transform.LayerNorm.weight
    state_dict["mlm_head.ln.b"] = mlm_head.transform.LayerNorm.bias

    # The NSP head does not have an unembedding
    # so we are only using weights from the MLM head
    # Note: BERT uses tied embeddings
    state_dict["unembed.W_U"] = mlm_head.decoder.weight.T
    state_dict["unembed.b_U"] = mlm_head.decoder.bias

    nsp_head = bert.cls.seq_relationship
    state_dict["nsp_head.W"] = nsp_head.weight.T
    state_dict["nsp_head.b"] = nsp_head.bias

    return state_dict



from typing import cast

import einops
import torch

from transformer_lens.HookedTransformerConfig import HookedTransformerConfig


def convert_llama_weights(llama, cfg: HookedTransformerConfig):
    state_dict = {}

    state_dict["embed.W_E"] = llama.model.embed_tokens.weight

    # Some models with the Llama architecture use Grouped Query Attention, and so for these we need to modify
    # the state dict keys for the K/V attention weight/biases, prepending "_" to the key names.
    using_gqa = cfg.n_key_value_heads is not None
    gqa_uscore = "_" if using_gqa else ""
    # need a cast since MyPy isn't smart enough to realize that using_gqa implies n_key_value_heads is not None
    n_kv_heads = cast(int, cfg.n_key_value_heads if using_gqa else cfg.n_heads)

    # llama has no biases anywhere and deals with everything else roughly like
    # GPTNeoX with different names

    assert cfg.d_mlp is not None  # keep mypy happy

    for l in range(cfg.n_layers):
        state_dict[f"blocks.{l}.ln1.w"] = llama.model.layers[l].input_layernorm.weight

        W_Q = llama.model.layers[l].self_attn.q_proj.weight
        W_K = llama.model.layers[l].self_attn.k_proj.weight
        W_V = llama.model.layers[l].self_attn.v_proj.weight

        # in case of quantization,
        # parameters should stay as bitsandbytes.nn.modules.Params4bit
        if not cfg.load_in_4bit:
            W_Q = einops.rearrange(W_Q, "(n h) m->n m h", n=cfg.n_heads)
            W_K = einops.rearrange(W_K, "(n h) m->n m h", n=n_kv_heads)
            W_V = einops.rearrange(W_V, "(n h) m->n m h", n=n_kv_heads)

        state_dict[f"blocks.{l}.attn.W_Q"] = W_Q
        state_dict[f"blocks.{l}.attn.{gqa_uscore}W_K"] = W_K
        state_dict[f"blocks.{l}.attn.{gqa_uscore}W_V"] = W_V

        state_dict[f"blocks.{l}.attn.b_Q"] = torch.zeros(
            cfg.n_heads, cfg.d_head, dtype=cfg.dtype, device=cfg.device
        )
        state_dict[f"blocks.{l}.attn.{gqa_uscore}b_K"] = torch.zeros(
            n_kv_heads,
            cfg.d_head,
            dtype=cfg.dtype,
            device=cfg.device,
        )
        state_dict[f"blocks.{l}.attn.{gqa_uscore}b_V"] = torch.zeros(
            n_kv_heads,
            cfg.d_head,
            dtype=cfg.dtype,
            device=cfg.device,
        )

        W_O = llama.model.layers[l].self_attn.o_proj.weight

        if not cfg.load_in_4bit:
            W_O = einops.rearrange(W_O, "m (n h)->n h m", n=cfg.n_heads)

        state_dict[f"blocks.{l}.attn.W_O"] = W_O.to(device=cfg.device)

        state_dict[f"blocks.{l}.attn.b_O"] = torch.zeros(
            cfg.d_model, dtype=cfg.dtype, device=cfg.device
        )

        state_dict[f"blocks.{l}.ln2.w"] = llama.model.layers[l].post_attention_layernorm.weight

        # in case of quantization,
        # parameters should stay as bitsandbytes.nn.modules.Params4bit
        if not cfg.load_in_4bit:
            state_dict[f"blocks.{l}.mlp.W_in"] = llama.model.layers[l].mlp.up_proj.weight.T
            state_dict[f"blocks.{l}.mlp.W_gate"] = llama.model.layers[l].mlp.gate_proj.weight.T
            state_dict[f"blocks.{l}.mlp.W_out"] = llama.model.layers[l].mlp.down_proj.weight.T
        else:
            state_dict[f"blocks.{l}.mlp.W_in"] = llama.model.layers[l].mlp.up_proj.weight
            state_dict[f"blocks.{l}.mlp.W_gate"] = llama.model.layers[l].mlp.gate_proj.weight
            state_dict[f"blocks.{l}.mlp.W_out"] = llama.model.layers[l].mlp.down_proj.weight

        state_dict[f"blocks.{l}.mlp.b_in"] = torch.zeros(
            cfg.d_mlp, dtype=cfg.dtype, device=cfg.device
        )
        state_dict[f"blocks.{l}.mlp.b_out"] = torch.zeros(
            cfg.d_model, dtype=cfg.dtype, device=cfg.device
        )

    state_dict["ln_final.w"] = llama.model.norm.weight

    state_dict["unembed.W_U"] = llama.lm_head.weight.T
    state_dict["unembed.b_U"] = torch.zeros(cfg.d_vocab, dtype=cfg.dtype, device=cfg.device)

    return state_dict



"""Hooked Transformer Token Typed Embed Component.

This module contains all the component :class:`TokenTypeEmbed`.
"""
from typing import Dict, Union

import torch
import torch.nn as nn
from jaxtyping import Int

from transformer_lens.HookedTransformerConfig import HookedTransformerConfig


class TokenTypeEmbed(nn.Module):
    """
    The token-type embed is a binary ids indicating whether a token belongs to sequence A or B. For example, for two sentences: "[CLS] Sentence A [SEP] Sentence B [SEP]", token_type_ids would be [0, 0, ..., 0, 1, ..., 1, 1]. `0` represents tokens from Sentence A, `1` from Sentence B. If not provided, BERT assumes a single sequence input. Typically, shape is (batch_size, sequence_length).

    See the BERT paper for more information: https://arxiv.org/pdf/1810.04805.pdf
    """

    def __init__(self, cfg: Union[Dict, HookedTransformerConfig]):
        super().__init__()
        self.cfg = HookedTransformerConfig.unwrap(cfg)
        self.W_token_type = nn.Parameter(torch.empty(2, self.cfg.d_model, dtype=self.cfg.dtype))

    def forward(self, token_type_ids: Int[torch.Tensor, "batch pos"]):
        return self.W_token_type[token_type_ids, :]



from typing import List, Callable
import argparse
from .data_classes import PruningConfig

def is_true(arg):
    if isinstance(arg, bool):
        return arg
    if arg.lower() in ['true', 't', '1']:
        return True
    return False

def split_list(arg):
    if isinstance(arg, list) or isinstance(arg, tuple):
        return arg
    return arg.split(',')


def cli_parser(
        c: PruningConfig,
        add_args_fn: Callable = None,
        add_args_exclude: List[str] = None,
    ):
    """
    Args:
        c (PruningConfig): default pruning config data class
        add_args_fn (Callable[parser] -> None): function to add additional
            arguments to the parser )
        add_arguments_exclude (List[str]): argument strings to manually exclude

    Returns:
        c (PruningConfig): updated pruning config data class
        args (argparse.Namespace): parsed arguments
    """
    # Build a CLI parser
    parser = argparse.ArgumentParser()

    if add_args_exclude is None:
        add_args_exclude = []

    if add_args_fn is not None:
        add_args_fn(parser)

    parser.add_argument('model_repo', type=str)
    parser.add_argument('-n', "--name", type=str, default=None, help="wandb run name")
    parser.add_argument('-r', '--reverse', action='store_true', help="cripple <--> focus")
    parser.add_argument('--n_steps', type=int, default=None)
    parser.add_argument('--model_device', type=str, default=None)

    args_exclude = [
        "model_repo",
        "name",
        "wandb_run_name",
        "reverse",
        "n_steps",
        "model_device",
#        "additional_datasets",
        *add_args_exclude,
    ]
    for key, val in c.arg_items(exclude=args_exclude):
        _type = type(val)
        _type = _type if not isinstance(val, bool) else str
        _type = _type if not isinstance(val, tuple) else str
        parser.add_argument(f'--{key}', type=_type, default=val)

    # Parse the argument
    args = parser.parse_args()

    c.model_repo = args.model_repo
    c.model_device = args.model_device
    c.wandb_run_name = args.name
    for key in c.arg_keys(exclude=args_exclude):
        if isinstance(c[key], bool):
            c[key] = is_true(getattr(args, key))
            continue
        if isinstance(c[key], tuple):
            c[key] = split_list(getattr(args, key))
            continue
        c[key] = getattr(args, key)
    if args.reverse:
        c.focus, c.cripple = c.cripple, c.focus
    # First do some pruning of the feed forward layers
    n_steps = args.n_steps
    if n_steps is None:
        n_steps = int( 1 / max(c.ff_frac, c.attn_frac) )
    c.n_steps = n_steps

    return c, args
